import sys, os, logging
import mob_suite.mob_recon
#test all mob_recon functions including aggregation of results
TEST_ROOT = os.path.dirname(__file__)
ROOT_MODULE = os.path.dirname(mob_suite.__file__)
logger=logging.getLogger()
LOG_FORMAT = '%(asctime)s %(levelname)s: %(message)s [in %(pathname)s:%(lineno)d]'
logging.basicConfig(format=LOG_FORMAT, level=logging.DEBUG)

if not os.path.exists(os.path.join(TEST_ROOT,"run_test")):
    os.mkdir(os.path.join(TEST_ROOT,"run_test"))


def test_mob_recon_with_mob_typer_report():
    if os.path.exists("run_test") == False:
        os.mkdir("run_test")
    #IncFIB,IncFII multi-plasmids
    args = [
        "--infile", os.path.dirname(__file__) + "/TestData/Pseudomonas/test_contigs.fasta",
        "--outdir", os.path.dirname(__file__)+"/run_test/mob_recon",
        "--debug",
        "--force"
    ]
    sys.argv[1:] = args
    mob_suite.mob_recon.main()

    mobtyper_results_file = os.path.join(TEST_ROOT,"run_test/mob_recon/mobtyper_results.txt")
    assert sum(1 for line in open(mobtyper_results_file)) == 2 , "Results file is empty, something went wrong"

def test_mob_recon_no_plasmid_biomarkers():

    args = [
        "--infile", ROOT_MODULE + "/example/assembly_no_biomarkers.fasta",
        "--outdir", TEST_ROOT + "/run_test/mob_recon_no_plasmid_markers",
        "--debug",
        "--force"
    ]

    sys.argv[1:] = args
    mob_suite.mob_recon.main()
    assert os.path.exists(os.path.join(TEST_ROOT,"run_test/mob_recon_no_plasmid_markers/biomarker_report.txt")) == False

def test_mob_recon_typical_run():
    args = [
        "--infile", ROOT_MODULE + "/example/SRR3703080_illumina_unicycler.fasta",
        "--outdir", TEST_ROOT+"/run_test/mob_recon_SRR3703080",
        "--debug",
        "--force"
    ]
    sys.argv[1:] = args
    mob_suite.mob_recon.main()

    assert os.path.exists(os.path.join(TEST_ROOT,"run_test/mob_recon_SRR3703080/biomarkers.blast.txt")) == True, "File does not exist"
    assert os.path.exists(os.path.join(TEST_ROOT,"run_test/mob_recon_SRR3703080/chromosome.fasta")) == True, "File does not exist"
    assert os.path.exists(os.path.join(TEST_ROOT,"run_test/mob_recon_SRR3703080/mobtyper_results.txt")) == True, "File does not exist"
    assert os.path.exists(os.path.join(TEST_ROOT,"run_test/mob_recon_SRR3703080/contig_report.txt")) == True, "File does not exits"
    assert os.path.exists(os.path.join(TEST_ROOT,"run_test/mob_recon_SRR3703080/plasmid_AA474.fasta")) == True, "File does not exists"

    


import os

MOB_CLUSTER_INFO_HEADER = [
    'id',
    'size',
    'gc',
    'md5',
    'organism',
    'primary_cluster_id',
    'primary_dist',
    'secondary_cluster_id',
    'secondary_dist',
    "rep_type(s)",
    "rep_type_accession(s)",
    "relaxase_type(s)",
    "relaxase_type_accession(s)"
]

MOB_TYPER_REPORT_HEADER = [
    'sample_id',
    'num_contigs',
    'size',
    'gc',
    'md5',
    'rep_type(s)',
    'rep_type_accession(s)',
    'relaxase_type(s)',
    'relaxase_type_accession(s)',
    'mpf_type',
    'mpf_type_accession(s)',
    'orit_type(s)',
    'orit_accession(s)',
    'predicted_mobility',
    'mash_nearest_neighbor',
    'mash_neighbor_distance',
    'mash_neighbor_identification',
    'primary_cluster_id',
    'secondary_cluster_id',
    'predicted_host_range_overall_rank',
    'predicted_host_range_overall_name',
    'observed_host_range_ncbi_rank',
    'observed_host_range_ncbi_name',
    'reported_host_range_lit_rank',
    'reported_host_range_lit_name',
    'associated_pmid(s)'
]

MOB_RECON_INFO_HEADER = [
    'sample_id',
    'molecule_type',
    'primary_cluster_id',
    'secondary_cluster_id',
    'contig_id',
    'size',
    'gc',
    'md5',
    'circularity_status',
    'rep_type(s)',
    'rep_type_accession(s)',
    'relaxase_type(s)',
    'relaxase_type_accession(s)',
    'mpf_type',
    'mpf_type_accession(s)',
    'orit_type(s)',
    'orit_accession(s)',
    'predicted_mobility',
    'mash_nearest_neighbor',
    'mash_neighbor_distance',
    'mash_neighbor_identification',
    'repetitive_dna_id',
    'repetitive_dna_type',
    'filtering_reason',

]

MGE_INFO_HEADER = [
    'sample_id',
    'molecule_type',
    'primary_cluster_id',
    'secondary_cluster_id',
    'contig_id',
    'size',
    'gc',
    'md5',
    'mge_id',
    'mge_acs',
    'mge_type',
    'mge_subtype',
    'mge_length',
    'mge_start',
    'mge_end',
    'contig_start',
    'contig_end',
    'length',
    'sstrand',
    'qcovhsp',
    'pident',
    'evalue',
    'bitscore'

]

NCBI_PLASMID_TAXONOMY_HEADER = [
    'sample_id',
    'num_contigs',
    'total_length',
    'gc',
    'md5',
    'rep_type(s)',
    'rep_type_accession(s)',
    'relaxase_type(s)',
    'relaxase_type_accession(s)',
    'mpf_type',
    'mpf_type_accession(s)',
    'orit_type(s)',
    'orit_accession(s)',
    'predicted_mobility',
    'mash_nearest_neighbor',
    'mash_neighbor_distance',
    'primary_cluster_id',
    'secondary_cluster_id',
    'organism',
    'taxid'
]

MOB_CLUSTER_INFO_HEADER = [
    'sample_id',
    'size',
    'gc',
    'md5',
    'organism',
    'taxid',
    'rep_type(s)',
    'rep_type_accession(s)',
    'relaxase_type(s)',
    'relaxase_type_accession(s)',
    'mpf_type',
    'mpf_type_accession(s)',
    'orit_type(s)',
    'orit_accession(s)',
    'predicted_mobility',
    'primary_cluster_id',
    'primary_dist',
    'secondary_cluster_id',
    'secondary_dist',
]

ACS_LETTER_VALUES = {
    'A': 0, 'B': 1, 'C': 2, 'D': 3, 'E': 4, 'F': 5, 'G': 6, 'H': 7, 'I': 8, 'J': 9,
    'K': 10, 'L': 11, 'M': 12, 'N': 13, 'O': 14, 'P': 15, 'Q': 16, 'R': 17, 'S': 18,
    'T': 19, 'U': 20, 'V': 21, 'W': 22, 'X': 23, 'Y': 24, 'Z': 25
}

ACS_VALUES_TO_LETTERS = {
    0: 'A', 1: 'B', 2: 'C', 3: 'D', 4: 'E', 5: 'F', 6: 'G', 7: 'H', 8: 'I', 9: 'J',
    10: 'K', 11: 'L', 12: 'M', 13: 'N', 14: 'O', 15: 'P', 16: 'Q', 17: 'R', 18: 'S',
    19: 'T', 20: 'U', 21: 'V', 22: 'W', 23: 'X', 24: 'Y', 25: 'Z'
}

MAX_ACS_VALUE = 650999
ACS_FORMAT_VALUES = [26000, 1000, 1]

LOG_FORMAT = '%(asctime)s %(levelname)s: %(message)s [in %(pathname)s:%(lineno)d]'

LIT_PLASMID_TAXONOMY_HEADER = [
    'sample_id', 'rep_type(s)', 'length', 'host_species', 'host_taxid', 'reported_host_range_taxid', 'pmid',
    'pmcid', 'doi', 'year', 'author', 'notes'
]



default_database_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'databases')

LOG_FORMAT = '%(asctime)s %(name)s %(levelname)s: %(message)s [in %(pathname)s:%(lineno)d]'



from Bio import SeqIO
from Bio.SeqUtils import gc_fraction
from mob_suite.blast import BlastRunner
from mob_suite.blast import BlastReader
import os, re, time
from subprocess import Popen, PIPE
import shutil, sys, logging, hashlib, random
import pandas as pd
from collections import OrderedDict
from operator import itemgetter
from ete3 import NCBITaxa
from mob_suite.constants import \
    MOB_TYPER_REPORT_HEADER, \
    MGE_INFO_HEADER


def getAssocValues(query_list_values, look_up_key, value_key, data):
    values = []

    for q in query_list_values:
        values.extend(get_data_associated_with_key(look_up_key, q, value_key, data))
    return values


def parseMashScreen(mash_results):
    mash_results = mash_results.decode('utf-8').split("\n")

    hits = {}

    for line in mash_results:
        if len(line) < 4:
            continue
        row = line.strip("\n").split("\t")

        ref_id = str(row[4])
        score = float(row[0])

        if ref_id not in hits:
            hits[ref_id] = {}

        hits[ref_id] = score

    return hits


def parseMash(mash_results):
    mash_results = mash_results.decode('utf-8').split("\n")

    hits = {}

    for line in mash_results:
        row = line.strip("\n").split("\t")
        if len(row) < 4:
            continue
        ref_id = str(row[0])
        query_id = str(row[1])
        score = float(row[2])

        if query_id not in hits:
            hits[query_id] = {}

        hits[query_id][ref_id] = score

    return hits


def filter_invalid_taxids(taxids):
    filtered = []
    for i in taxids:
        if i == 'NaN' or i == 'nan' or i == None or i == '-':
            continue

        try:
            filtered.append(int(i))

        except ValueError:
            continue

    return filtered


def getHeirarchy(taxid, ETE3DBTAXAFILE, database_directory):
    if not isETE3DBTAXAFILEexists(ETE3DBTAXAFILE):
        logging.info("Did not find taxa.sqlite in {}. Initializaing ete3 taxonomy database".format(ETE3DBTAXAFILE))
        initETE3Database(database_directory, ETE3DBTAXAFILE)

    ncbi = NCBITaxa(dbfile=ETE3DBTAXAFILE)
    if not isETE3DBTAXAFILEexists(ETE3DBTAXAFILE):
        logging.error(
            "Tried ete3 init, but still was not able to find taxa.sqlite file for ete3 lib in {}. Aborting".format(
                ETE3DBTAXAFILE))
        return ['-', '-']

    if not isinstance(taxid, int):
        return {'names': [], 'ranks': []}

    lineage = ncbi.get_lineage(taxid)
    names = ncbi.get_taxid_translator(lineage)
    ranks = []
    for id in lineage:
        ranks.append(ncbi.get_rank(id))

    return {'names': names, 'ranks': names}


def getTaxid(taxon, ETE3DBTAXAFILE, database_directory):
    if not isETE3DBTAXAFILEexists(ETE3DBTAXAFILE):
        logging.info("Did not find taxa.sqlite in {}. Initializaing ete3 taxonomy database".format(ETE3DBTAXAFILE))
        initETE3Database(database_directory, ETE3DBTAXAFILE)

    ncbi = NCBITaxa(dbfile=ETE3DBTAXAFILE)
    if not isETE3DBTAXAFILEexists(ETE3DBTAXAFILE):
        logging.error(
            "Tried ete3 init, but still was not able to find taxa.sqlite file for ete3 lib in {}. Aborting".format(
                ETE3DBTAXAFILE))
        return ['-', '-']

    taxid = ncbi.get_name_translator(taxon)

    return taxid



def NamesToTaxIDs(names, ETE3DBTAXAFILE, database_directory):
    if not isETE3DBTAXAFILEexists(ETE3DBTAXAFILE):
        logging.info("Did not find taxa.sqlite in {}. Initializaing ete3 taxonomy database".format(ETE3DBTAXAFILE))
        initETE3Database(database_directory, ETE3DBTAXAFILE)

    ncbi = NCBITaxa(dbfile=ETE3DBTAXAFILE)

    if not isETE3DBTAXAFILEexists(ETE3DBTAXAFILE):
        logging.error(
            "Tried ete3 init, but still was not able to find taxa.sqlite file for ete3 lib in {}. Aborting".format(
                ETE3DBTAXAFILE))
        return {}

    return ncbi.get_name_translator(names)



def getTaxonConvergence(taxids, ETE3DBTAXAFILE, database_directory):
    if not isETE3DBTAXAFILEexists(ETE3DBTAXAFILE):
        logging.info("Did not find taxa.sqlite in {}. Initializaing ete3 taxonomy database".format(ETE3DBTAXAFILE))
        initETE3Database(database_directory, ETE3DBTAXAFILE)

    ncbi = NCBITaxa(dbfile=ETE3DBTAXAFILE)

    if not isETE3DBTAXAFILEexists(ETE3DBTAXAFILE):
        logging.error(
            "Tried ete3 init, but still was not able to find taxa.sqlite file for ete3 lib in {}. Aborting".format(
                ETE3DBTAXAFILE))
        return ['-', '-']

    tax_ranks = ['genus', 'family', 'order', 'class', 'phylum', 'species']

    phylla = []
    lineages = []
    if len(taxids) == 0:
        return ['-', '-']

    elif len(taxids) == 1:
        common_taxids = ncbi.get_lineage(taxids[0])

    else:
        for taxid in taxids:
            lineage = ncbi.get_lineage(taxid)
            ranks = ncbi.get_rank(lineage)
            names = ncbi.get_taxid_translator(lineage)
            for id in ranks:
                rank = ranks[id]
                name = names[id]
                if rank == 'phylum':
                    phylla.append(name)

            lineages.append(lineage)

        common_taxids = lineages[0]
        for i in range(1, len(lineages)):
            common_taxids = list(set(common_taxids) - set(set(common_taxids) - set(lineages[i])))

    if len(common_taxids) == 0:
        return ['-', '-']

    taxa = {}

    for i in range(0, len(common_taxids)):
        rank = ncbi.get_rank([common_taxids[i]])[common_taxids[i]]
        name = ncbi.get_taxid_translator([common_taxids[i]])[common_taxids[i]]

        if not rank in tax_ranks:
            continue

        taxa[rank] = name

    for rank in tax_ranks:
        if rank in taxa:
            return [rank, taxa[rank]]

    phylla = list(set(phylla))
    sorted(phylla)
    if len(phylla) > 1:
        return ('multi-phylla', ",".join(phylla))
    elif len(phylla) == 1:
        return ('phylum', ",".join(phylla))

    return (['-', '-'])


def hostrange(replion_types, relaxase_types, mob_cluster_id, ncbi, lit, ETE3DBTAXAFILE, database_directory):
    host_range_predictions = {
        'observed_host_range_ncbi_name': '',
        'observed_host_range_ncbi_rank': '',
        'reported_host_range_lit_name': '',
        'reported_host_range_lit_rank': '',
        'predicted_host_range_overall_name': '',
        'predicted_host_range_overall_rank': '',
        'associated_pmid(s)': '-'

    }

    # Determine taxids associated with NCBI records
    if mob_cluster_id != '-':
        ncbi_cluster_taxids = getAssocValues([mob_cluster_id], 'primary_cluster_id', 'taxid', ncbi)

    else:
        ncbi_cluster_taxids = []

    if len(replion_types) > 0:
        ncbi_replicon_taxids = getAssocValues(replion_types, 'rep_type(s)', 'taxid', ncbi)
        lit_replicon_taxids = getAssocValues(replion_types, 'rep_type(s)', 'reported_host_range_taxid', lit)
        lit_replicon_taxids = lit_replicon_taxids + getAssocValues(replion_types, 'rep_type(s)', 'host_taxid', lit)
        pmids = list(set(getAssocValues(replion_types, 'rep_type(s)', 'pmid', lit)))
        sorted(pmids)
        pmids = filter_invalid_taxids(list(set(pmids)))
        host_range_predictions['associated_pmid(s)'] = '; '.join(str(x) for x in pmids)

    else:
        ncbi_replicon_taxids = []
        lit_replicon_taxids = []

    if len(relaxase_types) > 0:
        ncbi_relaxase_taxids = getAssocValues(relaxase_types, 'relaxase_type_accession(s)', 'taxid', ncbi)
    else:
        ncbi_relaxase_taxids = []

    if len(ncbi_cluster_taxids) == 0 and len(ncbi_replicon_taxids) == 0 and len(ncbi_relaxase_taxids) == 0:
        host_range_predictions['observed_host_range_ncbi_rank'] = '-'
        host_range_predictions['observed_host_range_ncbi_name'] = '-'
        ncbi_unique_taxids = []
    else:
        ncbi_unique_taxids = filter_invalid_taxids(
            list(set(ncbi_replicon_taxids + ncbi_cluster_taxids + ncbi_relaxase_taxids)))
        host_range_predictions['observed_host_range_ncbi_rank'], host_range_predictions[
            'observed_host_range_ncbi_name'] = getTaxonConvergence(ncbi_unique_taxids, ETE3DBTAXAFILE, database_directory)

    # Determine taxids associated with literature

    lit_unique_taxids = filter_invalid_taxids(list(set(lit_replicon_taxids)))

    host_range_predictions['reported_host_range_lit_rank'], host_range_predictions[
        'reported_host_range_lit_name'] = getTaxonConvergence(lit_unique_taxids, ETE3DBTAXAFILE, database_directory)

    # determine overall host range
    overall_taxids = filter_invalid_taxids(list(set(ncbi_unique_taxids + lit_unique_taxids)))
    host_range_predictions['predicted_host_range_overall_rank'], host_range_predictions[
        'predicted_host_range_overall_name'] = getTaxonConvergence(overall_taxids, ETE3DBTAXAFILE, database_directory)

    # move host-range prediction up to family when it is at genus or species level
    if host_range_predictions['predicted_host_range_overall_rank'] == 'genus' or host_range_predictions[
        'predicted_host_range_overall_rank'] == 'species':
        taxid = getTaxid(host_range_predictions['predicted_host_range_overall_name'], ETE3DBTAXAFILE, database_directory)
        heir = getHeirarchy(taxid, ETE3DBTAXAFILE, database_directory)
        names = heir['names']
        ranks = heir['ranks']

        for i in range(0, len(ranks)):
            if ranks[i] == 'Family':
                host_range_predictions['predicted_host_range_overall_rank'] = 'family'
                host_range_predictions['predicted_host_range_overall_name'] = names[i]

    unique_taxa =  sorted(list(set(host_range_predictions['observed_host_range_ncbi_name'].split(','))))
    taxa =  host_range_predictions['observed_host_range_ncbi_name'].split(",")
    ranks = host_range_predictions['observed_host_range_ncbi_rank'].split(",")
    unique_ranks = []
    for taxon in unique_taxa:
        for i in range(0,len(ranks)):
            if taxon == taxa[i]:
                unique_ranks.append(ranks[i])
                break

    host_range_predictions['observed_host_range_overall_rank'] = ",".join(unique_ranks)
    host_range_predictions['observed_host_range_ncbi_name'] = ",".join(unique_taxa)

    unique_taxa =  sorted(list(set(host_range_predictions['predicted_host_range_overall_name'].split(','))))
    taxa =  host_range_predictions['predicted_host_range_overall_name'].split(",")
    ranks = host_range_predictions['predicted_host_range_overall_rank'].split(",")
    unique_ranks = []
    for taxon in unique_taxa:
        for i in range(0,len(ranks)):
            if taxon == taxa[i]:
                unique_ranks.append(ranks[i])
                break
    host_range_predictions['predicted_host_range_overall_rank'] = ",".join(unique_ranks)
    host_range_predictions['predicted_host_range_overall_name'] = ",".join(unique_taxa)


    return host_range_predictions


def blastn(input_fasta, blastdb, min_ident, min_cov, evalue, min_length, out_dir, blast_results_file, logging,
           seq_filterfile=None, num_threads=1, max_length=400000, min_hsp_cov=1):
    blast_runner = BlastRunner(input_fasta, out_dir)
    blast_runner.run_blast(query_fasta_path=input_fasta, blast_task='megablast', db_path=blastdb,
                           db_type='nucl', min_cov=min_cov, min_ident=min_ident, evalue=evalue,
                           blast_outfile=blast_results_file, logging=logging, num_threads=num_threads, word_size=11,
                           seq_id_file=seq_filterfile)

    if os.path.getsize(blast_results_file) == 0:
        os.remove(blast_results_file)
        return False

    blast_df = BlastReader(blast_results_file, logging).df
    blast_df['length'] = blast_df['length'].astype('int64')
    blast_df['qlen'] = blast_df['qlen'].astype('int64')
    blast_df['qcovs'] = blast_df['qcovs'].astype('float64')
    blast_df['qcovhsp'] = blast_df['qcovhsp'].astype('float64')
    blast_df['evalue'] = blast_df['evalue'].astype('float64')
    blast_df['pident'] = blast_df['pident'].astype('float64')
    blast_df = blast_df.loc[blast_df['length'] >= min_length]
    blast_df = blast_df.loc[blast_df['qlen'] <= max_length]
    blast_df = blast_df.loc[blast_df['qcovs'] >= min_cov]
    blast_df = blast_df.loc[blast_df['qcovhsp'] >= min_hsp_cov]
    blast_df = blast_df.loc[blast_df['evalue'] <= evalue]
    blast_df = blast_df.loc[blast_df['pident'] >= min_ident]

    blast_df = blast_df.reset_index(drop=True)
    blast_df = fixStart(blast_df)
    blast_df.to_csv(blast_results_file, sep='\t', header=True, line_terminator='\n', index=False)

    return True


def tblastn(input_fasta, blastdb, min_ident, min_covs, evalue, out_dir, blast_results_file, logging, num_threads=1,
            min_covhsp=25, seq_id_file=None):
    blast_runner = BlastRunner(input_fasta, out_dir)

    blast_runner.run_tblastn(query_fasta_path=input_fasta, blast_task='megablast', db_path=blastdb,
                             db_type='protein', min_cov=min_covs, min_ident=min_ident, evalue=evalue,
                             blast_outfile=blast_results_file,
                             num_threads=num_threads, seq_id_file=seq_id_file, logging=logging)

    if os.path.getsize(blast_results_file) == 0:
        os.remove(blast_results_file)
        return False

    blast_df = BlastReader(blast_results_file, logging).df

    blast_df = blast_df.loc[blast_df['pident'] >= min_ident]
    blast_df = blast_df.loc[blast_df['qcovs'] >= min_covs]
    blast_df = blast_df.loc[blast_df['qcovhsp'] >= min_covhsp]
    blast_df = blast_df.loc[blast_df['evalue'] <= evalue]
    blast_df = fixStart(blast_df)
    blast_df = blast_df.sort_values(['sseqid', 'sstart', 'send', 'bitscore'], ascending=[True, True, True, False])
    blast_df = blast_df.reset_index(drop=True)
    blast_df.to_csv(blast_results_file, sep='\t', header=True, line_terminator='\n', index=False)

    return True


def isETE3DBTAXAFILEexists(ETE3DBTAXAFILE):
    if not os.path.exists(ETE3DBTAXAFILE):
        return False
    else:
        return True


def initETE3Database(database_directory, ETE3DBTAXAFILE):
    lockfilepath = os.path.join(database_directory, ".lock")

    if os.path.exists(lockfilepath) == False:
        open(file=lockfilepath, mode="w").close()
        logging.info("Placed lock file at {}".format(lockfilepath))
    else:
        while os.path.exists(lockfilepath):
            elapsed_time = time.time() - os.path.getmtime(lockfilepath)
            logging.info("Lock file found at {}. Waiting for other processes to finish ete3 database init ...".format(
                lockfilepath))
            logging.info(
                "Elapsed time {} min. Will continue processing after 16 min mark.".format(int(elapsed_time / 60)))
            if elapsed_time >= 1000:
                logging.info(
                    "Elapsed time {} min. Assuming previous process completed all init steps. Continue ...".format(
                        int(elapsed_time / 60)))
                try:  # if previous process failed, no processes are running and > 16 min passed since the lock was created
                    os.remove(lockfilepath)
                except:  # continue if file was removed by other process
                    pass
                break
            time.sleep(60)  # recheck every 1 min if lock file was removed by other process
        logging.info("Lock file no longer exists. Assuming init process completed successfully")

    ncbi = NCBITaxa()
    ncbi.dbfile = ETE3DBTAXAFILE
    ncbi.update_taxonomy_database()

    try:
        os.remove(lockfilepath)
        logging.info("Lock file removed.")
    except:
        logging.warning("Lock file is already removed by some other process.")
        pass

    try:
        os.remove(os.path.join(os.getcwd(), "taxdump.tar.gz"))
        logging.info("Removed residual taxdump.tar.gz as ete3 is not doing proper cleaning job.")
    except:
        pass
    logging.info("ETE3 database init completed successfully.")



def ETE3_db_status_check(taxid, lockfilepath, ETE3DBTAXAFILE, logging):
    """
    Place a lock file while using ETE3 taxonomy database (taxa.sqlite) to prevent accidental concurrent multiprocess update
    Parameters:
        taxid - the taxonomy id which is 1 by default for database health testing
        lockfilepath - path to the database lock file
        ETE3DBTAXAFILE - path to ETE3 taxa.sqlite file
        logging - logger object for logging messages
    Returns:
        Bool: True/False value with regards to database usage.
              If .lock file is not removed after 10 min, program exits
    """
    max_time = 600
    elapsed_time = 0

    while os.path.exists(lockfilepath) == True and elapsed_time < max_time:
        interval = int(random.randrange(10, 60))
        logging.info(
            "Taxonomy lock file {} exists..waiting {}s for other process to complete".format(lockfilepath, interval))
        time.sleep(interval)
        elapsed_time += interval

    if os.path.exists(lockfilepath) == True:
        # delete lock file older than 1 hour
        modTimesinceEpoc = os.path.getmtime(lockfilepath)
        elapsedTime = (time.time() - modTimesinceEpoc) / 60
        if elapsedTime >= 10:
            os.remove(lockfilepath)
        else:
            logging.error(
                "Taxonomy lock file {} still exists after several attempts. Please delete lock file to continue".format(
                    lockfilepath))
        return False

    else:
        logging.info("Creating Lock file {}".format(lockfilepath))

        #some file systems are read-only which will not support lock file writting
        try:
            open(file=lockfilepath, mode="w").close()
        except Exception as e:
            logging.info(e)
            pass

        logging.info("Testing ETE3 taxonomy db {}".format(ETE3DBTAXAFILE))
        ncbi = NCBITaxa(dbfile=ETE3DBTAXAFILE)

        lineage = ncbi.get_lineage(taxid)

        try:
            os.remove(lockfilepath)
            logging.info("Lock file removed.")
        except Exception as e:
            logging.warning("Lock file is already removed by some other process or read-only file system")
            logging.warning(e)

        if len(lineage) > 0:
            return True
        else:
            return False


'''
INPUT: Read file into pandas df
Returns: list of lines in file with a dict of key value pairs from the header
'''


def read_file_to_dict(file, header, separater="\t"):
    data = pd.read_csv(file, sep=separater, header=0, names=header, encoding="UTF-8")
    records = []
    header_len = len(header)
    for index, row in data.iterrows():
        report = {}
        for i in range(0, header_len):
            name = header[i]
            if name in row:
                report[name] = row[name]
            else:
                report[name] = '-'
        records.append(report)
    return records


'''
    Input: Dictionary with format key :{dict}
    Output: Dictionary indexed by alternate key specified by user
'''


def dict_from_alt_key(dictionary, new_key):
    new_dict = {}
    for id in dictionary:
        values = dictionary[id]
        if new_key in values:
            if not values[new_key] in new_dict:
                new_dict[values[new_key]] = {}
            new_dict[values[new_key]][id] = values
    return new_dict


'''
    Input: Dictionary with format key :{dict}
    Output: Dictionary indexed by alternate key specified by user
'''


def dict_from_alt_key_list(data, new_key):
    new_dict = {}

    for d in data:
        if new_key in d:
            if not d[new_key] in new_dict:
                new_dict[d[new_key]] = {}
            new_dict[d[new_key]] = d

    return new_dict


'''

'''


def get_data_associated_with_key(look_up_key, look_up_value, value_key, dictionary):
    asc_values = []

    for id in dictionary:
        values = dictionary[id]
        if look_up_key in values:
            terms = values[look_up_key].split(',')
            if not look_up_value in terms:
                continue
            if value_key in values and str(values[value_key]) != 'nan':
                asc_values.append(values[value_key])

    return asc_values


'''
    Input: Path to TSV file with MOB_CLUSTER_INFO_HEADER fields as the header lines
    Output: Dictionary of sequence indexed by sequence identifier
'''


def read_sequence_info(file, header):
    if os.path.getsize(file) == 0:
        return dict()
    data = pd.read_csv(file, sep='\t', header=0, names=header)
    sequences = dict()
    for index, row in data.iterrows():
        sample_id = row['sample_id']
        sequences[sample_id] = {}
        for i in range(0, len(header)):
            if header[i] == 'id':
                v = index
            else:
                if str(row[header[i]]) == 'nan':
                    v = '-'
                else:
                    v = row[header[i]]
            sequences[sample_id][header[i]] = v

    return sequences


def check_dependencies(logger):
    external_programs = ['blastn', 'makeblastdb', 'tblastn']
    missing = 0
    for program in external_programs:
        path = shutil.which(program)
        if path is None:
            missing += 1
            logger.error("ERROR: Missing program: {}".format(program, ))
        else:
            logger.info("SUCCESS: Found program {} at {}".format(program, path))
    if missing > 0:
        logger.error("Error, you are missing needed programs for mob-suite, please install them and retry")
        sys.exit(-1)


def calc_md5(seq):
    seq = str(seq).encode()
    md5 = hashlib.md5()
    md5.update(seq)
    return md5.hexdigest()


def fixStart(blast_df):
    for index, row in blast_df.iterrows():

        sstart = int(blast_df.at[index, 'sstart'])
        send = int(blast_df.at[index, 'send'])
        if send < sstart:
            temp = sstart
            blast_df.at[index, 'sstart'] = send
            blast_df.at[index, 'send'] = temp
        qstart = int(blast_df.at[index, 'qstart'])
        qend = int(blast_df.at[index, 'qend'])
        if qend < qstart:
            temp = qstart
            blast_df.at[index, 'qstart'] = qend
            blast_df.at[index, 'qend'] = temp

    blast_df = blast_df.astype({"qstart": 'int64', "qend": 'int64',
                                "sstart": 'int64', "send": 'int64',
                                "evalue": 'float', "bitscore": 'int64',
                                'length': 'int64', 'qlen': 'int64',
                                'slen': 'int64', 'mismatch': 'int64',
                                'pident': 'float', 'qcovs': 'float', 'qcovhsp': 'float'})

    return blast_df


def read_fasta_dict(fasta_file):
    seqs = dict()
    with open(fasta_file, "r") as handle:
        for record in SeqIO.parse(handle, "fasta"):
            seqs[str(record.id)] = str(record.seq)
    handle.close()
    return seqs


'''
    Input: Fasta file and out directory
    Output: writes individual files for each sequence record
'''


def write_individual_fasta(input_fasta, out_dir):
    with open(input_fasta, "r") as handle:
        for record in SeqIO.parse(handle, "fasta"):
            out_fasta = os.path.join(out_dir, "{}.fasta".format(record.id))
            with open(out_fasta, "w", encoding="utf-8") as fh:
                fh.write("\n>{}\n{}\n".format(record.id, record.seq))
                fh.close()


def write_fasta_dict(seqs, fasta_file):
    with open(fasta_file, "w") as handle:
        for id in seqs:
            handle.write(">{}\n{}\n".format(id, seqs[id]))
    handle.close()


def verify_init(logger, database_dir):
    mob_init_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'mob_init.py')
    status_file = os.path.join(database_dir, 'status.txt')
    if not os.path.isfile(status_file):
        logger.info('MOB-databases need to be initialized, this will take some time')
        p = Popen([sys.executable, mob_init_path, '-d', database_dir],
                  stdout=PIPE,
                  stderr=PIPE,
                  shell=False)

        stdout, stderr = p.communicate()
        return_code = p.returncode
        logger.info("Return code {}".format(return_code))

        # Verify if no errors were captured during the mob_init script run. Otherwise abort
        if len(re.findall("error", stderr.decode())) != 0 or return_code != 0:
            logger.error("Something went wrong with database download or unpacking"
                         "Check MOB-Suite databases directory and your Internet connection.")
            sys.exit(-1)


def remove_split_hits(blast_df, query_col, contig_id_col, contig_start_col, contig_end_col, bitscore_col):
    blast_df = blast_df.sort_values([query_col, contig_id_col, contig_start_col, contig_end_col, bitscore_col],
                                    ascending=[True, True, True, True, False])
    hits = {}
    for index, row in blast_df.iterrows():
        query_id = row[query_col]
        contig_id = row[contig_id_col]
        contig_start = int(row[contig_start_col])
        contig_end = int(row[contig_end_col])
        length = row['slen']
        score = row['bitscore']
        if contig_start == 1 or contig_end == length:

            if query_id not in hits:
                hits[query_id] = {}

            if contig_id not in hits[query_id]:
                hits[query_id][contig_id] = {'start': {}, 'end': {}}

            if contig_start == 1:
                hits[query_id][contig_id]['start'][index] = {'start': contig_start, 'end': contig_end, 'score': score}
            else:
                hits[query_id][contig_id]['end'][index] = {'start': contig_start, 'end': contig_end, 'score': score}

    filter = []

    for query_id in hits:
        for contig_id in hits[query_id]:
            positions = hits[query_id][contig_id]

            # skip records without both a begining and end hit
            if len(positions['start']) == 0 or len(positions['end']) == 0:
                continue

            start_keys = list(positions['start'].keys())
            end_keys = list(positions['end'].keys())
            max_bitscore = 0
            top_side = ''

            for index in start_keys:
                score = positions['start'][index]['score']
                if score > max_bitscore:
                    max_bitscore = score
                    top_side = 'start'

            for index in end_keys:
                score = positions['end'][index]['score']
                if score > max_bitscore:
                    max_bitscore = score
                    top_side = 'end'

            if top_side == 'start':
                filter += start_keys
            else:
                filter += end_keys

    blast_df = blast_df[~blast_df.index.isin(filter)]
    blast_df.reset_index(drop=True)
    return blast_df


def filter_overlaping_records(blast_df, overlap_threshold, contig_id_col, contig_start_col, contig_end_col,
                              bitscore_col):
    prev_contig_id = ''
    prev_index = -1
    prev_contig_start = -1
    prev_contig_end = -1
    prev_score = -1
    filter_indexes = list()
    exclude_filter = dict()

    for index, row in blast_df.iterrows():

        contig_id = row[contig_id_col]
        contig_start = row[contig_start_col]
        contig_end = row[contig_end_col]
        score = float(row[bitscore_col])


        if prev_contig_id == '':
            prev_index = index
            prev_contig_id = contig_id
            prev_contig_start = contig_start
            prev_contig_end = contig_end
            prev_score = score
            continue

        if contig_id != prev_contig_id:
            prev_index = index
            prev_contig_id = contig_id
            prev_contig_start = contig_start
            prev_contig_end = contig_end
            prev_score = score
            continue
        if prev_index in filter_indexes:
            continue
        if (contig_start >= prev_contig_start and contig_start <= prev_contig_end) or (
                contig_end >= prev_contig_start and contig_end <= prev_contig_end):
            overlap = abs(contig_start - prev_contig_end)


            if overlap > overlap_threshold:
                if prev_score < score:
                    filter_indexes.append(prev_index)
                else:
                    filter_indexes.append(index)

        prev_index = index
        prev_contig_id = contig_id
        prev_contig_start = contig_start
        prev_contig_end = contig_end
        prev_score = score

    for index in exclude_filter:
        filter_indexes.append(index)
    blast_df.drop(filter_indexes, inplace=True)

    return blast_df.reset_index(drop=True)


def recursive_filter_overlap_records(blast_df, overlap_threshold, contig_id_col, contig_start_col, contig_end_col,
                                     bitscore_col):
    size = len(blast_df)
    prev_size = 0

    while size != prev_size:
        blast_df = filter_overlaping_records(blast_df, overlap_threshold, contig_id_col, contig_start_col,
                                             contig_end_col, bitscore_col).sort_values(
            ['sseqid', 'sstart', 'send', 'bitscore'], ascending=[True, True, True, False])
        prev_size = size
        size = len(blast_df)
    return blast_df


def fix_fasta_header(in_fasta, out_fasta):
    ids = {}
    fh = open(out_fasta, 'w')
    incr=0
    with open(in_fasta, "r") as handle:
        for record in SeqIO.parse(handle, "fasta"):
            id = str(record.description)
            status = 'false'
            if 'circular=true' in id or '_circ' in id:
                status = 'true'
            seq = str(record.seq.upper())
            md5 = calc_md5(seq)
            new_id = "{}_{}_circular={}".format(incr,md5,status)
            ids[new_id] = id
            fh.write(">{}\n{}\n".format(new_id,seq))
            incr+=1
    handle.close()
    fh.close()
    return ids


''''
    Accepts fasta file and returns size, number of sequence records and gc %
'''


def calcFastaStatsIndividual(fasta):
    stats = {}
    for record in SeqIO.parse(fasta, "fasta"):
        id = record.id
        seq = record.seq
        genome_size = len(seq)
        gc = gc_fraction(seq)
        md5 = calc_md5(seq)
        stats[id] = {
            'total_length': genome_size,
            'gc': gc,
            'md5': md5
        }

    return stats


''''
    Accepts fasta file and returns size, number of sequence records and gc %
'''


def calcFastaStats(fasta):
    num_seqs = 0;
    seq = ''
    for record in SeqIO.parse(fasta, "fasta"):
        num_seqs += 1
        seq = seq + record.seq
    genome_size = len(seq)
    gc = gc_fraction(seq)
    md5 = calc_md5(seq)

    return {
        'num_seq': num_seqs,
        'size': genome_size,
        'gc_content': gc,
        'md5': md5
    }


def filterFastaByIDs(in_fasta, out_fasta, include_list):
    fh = open(out_fasta, 'w')

    for record in SeqIO.parse(in_fasta, "fasta"):
        if record.id in include_list:
            fh.write(">{}\n{}\n".format(record.id, record.seq))

    fh.close()


def init_console_logger(lvl=2):
    LOG_FORMAT = '%(asctime)s %(name)s %(levelname)s: %(message)s [in %(pathname)s:%(lineno)d]'

    logging_levels = [logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG]
    report_lvl = logging_levels[min(lvl, 3)]

    logging.basicConfig(format=LOG_FORMAT, level=report_lvl)
    return logging.getLogger(__name__)


def writeReport(data_list, header, outfile):
    with open(outfile, 'w') as fh:
        fh.write("{}\n".format("\t".join(header)))
        for i in range(0, len(data_list)):
            data = data_list[i]
            row = []
            for h in header:
                if h in data:
                    if data[h] != '':
                        row.append(str(data[h]))
                    else:
                        row.append('-')
                else:
                    row.append("-")
            fh.write("{}\n".format("\t".join(row)))
        fh.close()


def sort_biomarkers(biomarker_dict):
    for id in biomarker_dict:
        acs = biomarker_dict[id]['acs']
        types = biomarker_dict[id]['types']

        if len(acs) == 0 or not isinstance(acs, list):
            continue

        tmp_dict = {}

        for i in range(0, len(acs)):
            tmp_dict[acs[i]] = types[i]

        tmp_dict = OrderedDict(sorted(tmp_dict.items(), key=itemgetter(1), reverse=False))

        biomarker_dict[id]['acs'] = list(tmp_dict.keys())
        biomarker_dict[id]['types'] = list(tmp_dict.values())

    return biomarker_dict


def determine_mpf_type(hits):
    if len(hits) == 0 or isinstance(hits, str):
        return hits
    types = dict()
    for hit in hits:
        if not hit in types:
            types[hit] = 0
        types[hit] += 1

    return max(types, key=lambda i: types[i])


def build_mobtyper_report(plasmid_contig_info, out_dir, outfile, seq_dict, ncbi, lit, ETE3DBTAXAFILE, database_directory):
    mob_typer_results = {}
    for clust_id in plasmid_contig_info:

        cluster_file = open(os.path.join(out_dir, "plasmid_{}.fasta".format(clust_id)), 'w')
        logging.info(
            "Writting plasmid sequences to {}".format(os.path.join(out_dir, "plasmid_{}.fasta".format(clust_id))))
        if clust_id not in mob_typer_results:
            mob_typer_results[clust_id] = {}
        for field in MOB_TYPER_REPORT_HEADER:
            if not field in mob_typer_results[clust_id]:
                mob_typer_results[clust_id][field] = []

        data = plasmid_contig_info[clust_id]

        # Put contig report data into MOB-typer report header
        # aggregating the data by cluster id
        cluster_seq = []
        for contig_id in data:
            if contig_id in seq_dict:
                cluster_file.write(">{}\n{}\n".format(contig_id, seq_dict[contig_id]))
                cluster_seq.append(seq_dict[contig_id])
            for field in MOB_TYPER_REPORT_HEADER:
                if field in data[contig_id]:
                    if isinstance(data[contig_id][field], list) and len(data[contig_id][field]) == 0:
                        continue
                    if data[contig_id][field] == '':
                        continue
                    mob_typer_results[clust_id][field].append(data[contig_id][field])

        # overwrite individual seq stat calculations with the overall
        cluster_seq = sorted(cluster_seq,key=len)
        seq = "".join(cluster_seq)
        mob_typer_results[clust_id]['md5'] = [calc_md5(seq)]
        mob_typer_results[clust_id]['gc'] = [gc_fraction(seq)]
        mob_typer_results[clust_id]['size'] = [len(seq)]
        mob_typer_results[clust_id]['num_contigs'] = len(cluster_seq)

        # Sort MOB-typer biomarker results
        replicon = sort_biomarkers({'rep': {'types': mob_typer_results[clust_id]['rep_type(s)'],
                                            'acs': mob_typer_results[clust_id]['rep_type_accession(s)']}})
        mob_typer_results[clust_id]['rep_type(s)'] = ",".join(replicon['rep']['types'])
        mob_typer_results[clust_id]['rep_type_accession(s)'] = ",".join(replicon['rep']['acs'])

        relaxase = sort_biomarkers({'mob': {'types': mob_typer_results[clust_id]['relaxase_type(s)'],
                                            'acs': mob_typer_results[clust_id]['relaxase_type_accession(s)']}})
        mob_typer_results[clust_id]['relaxase_type(s)'] = ",".join(relaxase['mob']['types'])
        mob_typer_results[clust_id]['relaxase_type_accession(s)'] = ",".join(relaxase['mob']['acs'])

        if len(mob_typer_results[clust_id]['mpf_type']) > 0:
            if isinstance(mob_typer_results[clust_id]['mpf_type'], str):
                mob_typer_results[clust_id]['mpf_type'] = determine_mpf_type(
                    mob_typer_results[clust_id]['mpf_type'].split(','))
            else:
                tmp = []
                for i in range(0, len(mob_typer_results[clust_id]['mpf_type'])):
                    tmp += mob_typer_results[clust_id]['mpf_type'][i].split(',')
                mob_typer_results[clust_id]['mpf_type'] = determine_mpf_type(tmp)
                del (tmp)
        else:
            mob_typer_results[clust_id]['mpf_type'] = ''

        mob_typer_results[clust_id]['mpf_type_accession(s)'] = ",".join(
            mob_typer_results[clust_id]['mpf_type_accession(s)'])
        mob_typer_results[clust_id]['orit_type(s)'] = ",".join(mob_typer_results[clust_id]['orit_type(s)'])
        mob_typer_results[clust_id]['orit_accession(s)'] = ",".join(mob_typer_results[clust_id]['orit_accession(s)'])

        # Assign mobility
        mob_typer_results[clust_id]['predicted_mobility'] = 'non-mobilizable'
        if len(mob_typer_results[clust_id]['relaxase_type(s)']) > 0 and len(
                mob_typer_results[clust_id]['mpf_type']) > 0:
            mob_typer_results[clust_id]['predicted_mobility'] = 'conjugative'
        elif (len(mob_typer_results[clust_id]['relaxase_type(s)']) > 0 or len(
                mob_typer_results[clust_id]['orit_type(s)']) > 0):
            mob_typer_results[clust_id]['predicted_mobility'] = 'mobilizable'

        if isinstance(mob_typer_results[clust_id]['primary_cluster_id'], list) and len(
                mob_typer_results[clust_id]['primary_cluster_id']) > 0:
            mob_typer_results[clust_id]['primary_cluster_id'] = mob_typer_results[clust_id]['primary_cluster_id'][0]
            mob_typer_results[clust_id]['mash_nearest_neighbor'] = mob_typer_results[clust_id]['mash_nearest_neighbor'][
                0]
            mob_typer_results[clust_id]['mash_neighbor_distance'] = \
            mob_typer_results[clust_id]['mash_neighbor_distance'][0]
            mob_typer_results[clust_id]['mash_neighbor_identification'] = \
            mob_typer_results[clust_id]['mash_neighbor_identification'][0]

        if isinstance(mob_typer_results[clust_id]['sample_id'], list) and len(
                mob_typer_results[clust_id]['sample_id']) > 0:
            mob_typer_results[clust_id]['sample_id'] = "{}:{}".format(mob_typer_results[clust_id]['sample_id'][0],
                                                                      mob_typer_results[clust_id]['primary_cluster_id'])

        if isinstance(mob_typer_results[clust_id]['secondary_cluster_id'], list) and len(
                mob_typer_results[clust_id]['secondary_cluster_id']) > 0:
            mob_typer_results[clust_id]['secondary_cluster_id'] = mob_typer_results[clust_id]['secondary_cluster_id'][0]

        if mob_typer_results[clust_id]['num_contigs'] > 1:
            mob_typer_results[clust_id]['circularity_status'] = 'incomplete'

        if len(replicon['rep']['types']) > 0:
            rep_types = replicon['rep']['types']
        else:
            rep_types = []

        if len(relaxase['mob']['acs']) > 0:
            relaxase_types = relaxase['mob']['acs']
        else:
            relaxase_types = []

        if mob_typer_results[clust_id]['primary_cluster_id'] != '' and not 'novel_' in mob_typer_results[clust_id][
            'primary_cluster_id']:
            mob_cluster_id = mob_typer_results[clust_id]['primary_cluster_id']
        else:
            mob_cluster_id = '-'

        host_range = hostrange(rep_types, relaxase_types, mob_cluster_id, ncbi, lit, ETE3DBTAXAFILE, database_directory)

        for field in host_range:
            mob_typer_results[clust_id][field] = host_range[field]

        for element in MOB_TYPER_REPORT_HEADER:
            if element in mob_typer_results[clust_id]:
                if isinstance(mob_typer_results[clust_id][element], list):
                    mob_typer_results[clust_id][element] = ', '.join(
                        str(x) for x in mob_typer_results[clust_id][element])

    results = []
    for clust_id in mob_typer_results:
        results.append(mob_typer_results[clust_id])

    writeReport(results, MOB_TYPER_REPORT_HEADER, outfile)
    return


def add_biomarker_results(biomarker_df, df_column_name_biomarker, df_column_name_seqid, contig_info,
                          contig_info_type_key, contig_info_acs_key, type_col_num=1, type_acs_num=0, delimeter='|'):
    results = {}
    for index, row in biomarker_df.iterrows():

        biomarker = row[df_column_name_biomarker].split(delimeter)

        if type_col_num + 1 > len(biomarker):
            logging.error(
                "Specified column for biomarker type: {} does not exist in biomarker id field: {}".format(type_col_num,
                                                                                                          row[
                                                                                                              df_column_name_biomarker]))
            continue

        if type_acs_num + 1 > len(biomarker):
            logging.error("Specified column for biomarker type: {} does not exist in biomarker id field: {}".format(
                type_acs_num, row[df_column_name_biomarker]))
            continue

        seqid = row[df_column_name_seqid]
        if seqid[len(seqid) - 1] == '|':
            tmp = seqid.split('|')
            #
            # tmp[0] in ['gb','emb','ddj','ref'] and
            if len(tmp) > 1:
                seqid = tmp[1]

        if not seqid in results:
            results[seqid] = {'types': [], 'acs': []}

        results[seqid]['types'].append(biomarker[type_col_num])
        results[seqid]['acs'].append(biomarker[type_acs_num])

    results = sort_biomarkers(results)

    for seqid in results:
        if not seqid in contig_info:
            continue

        if contig_info_type_key in contig_info[seqid]:
            contig_info[seqid][contig_info_type_key] = ','.join(results[seqid]['types'])
        else:
            logging.error(
                "Error: {} contig_info_type_key not found in contig_info, check name and fix".format(
                    contig_info_type_key))

        if contig_info_acs_key in contig_info[seqid]:
            contig_info[seqid][contig_info_acs_key] = ','.join(results[seqid]['acs'])
        else:
            logging.error(
                "Error: {} contig_info_acs_key not found in contig_info, check name and fix".format(
                    contig_info_type_key))

    return contig_info


def identify_biomarkers(contig_info, fixed_fasta, tmp_dir, min_length, logging,
                        replicon_ref, min_rep_ident, min_rep_cov, min_rep_evalue, replicon_blast_results,
                        mob_ref, min_mob_ident, min_mob_cov, min_mob_evalue, mob_blast_results,
                        mpf_ref, min_mpf_ident, min_mpf_cov, min_mpf_evalue, mpf_blast_results,
                        repetitive_mask_file, min_rpp_ident, min_rpp_cov, min_rpp_evalue,
                        plasmid_orit, orit_blast_results, repetitive_blast_results,
                        num_threads=1):
    # blast replicon database
    logging.info("Blasting replicon sequences {} against {}".format(replicon_ref, fixed_fasta))
    blastn(input_fasta=replicon_ref, blastdb=fixed_fasta, min_ident=min_rep_ident, min_cov=min_rep_cov,
           evalue=min_rep_evalue, min_length=80, out_dir=tmp_dir,
           blast_results_file=replicon_blast_results, num_threads=num_threads, logging=logging, min_hsp_cov=30)

    logging.info("Filtering replicon blast results {} ".format(replicon_blast_results))
    rep_blast_df = BlastReader(replicon_blast_results, logging=logging).df
    if len(rep_blast_df) > 0:
        rep_blast_df = rep_blast_df.drop(0)
        rep_blast_df = fixStart(rep_blast_df)
        rep_blast_df = remove_split_hits(rep_blast_df, 'qseqid', 'sseqid', 'sstart', 'send', 'bitscore').sort_values(
            ['sseqid', 'sstart', 'send', 'bitscore'], ascending=[True, True, True, False])

        rep_blast_df = recursive_filter_overlap_records(rep_blast_df, 5, 'sseqid', 'sstart', 'send', 'bitscore')

        contig_info = add_biomarker_results(biomarker_df=rep_blast_df, df_column_name_biomarker='qseqid',
                                            df_column_name_seqid='sseqid', contig_info=contig_info,
                                            contig_info_type_key='rep_type(s)',
                                            contig_info_acs_key='rep_type_accession(s)', delimeter='|')

    del (rep_blast_df)

    # blast relaxase database
    logging.info("Blasting relaxase sequences {} against {}".format(mob_ref, fixed_fasta))
    tblastn(input_fasta=mob_ref, blastdb=fixed_fasta, min_ident=min_mob_ident, min_covs=min_mob_cov,
            evalue=min_mob_evalue, out_dir=tmp_dir, logging=logging,
            blast_results_file=mob_blast_results, num_threads=num_threads)

    logging.info("Filtering relaxase blast results {} ".format(mob_blast_results))

    mob_blast_df = BlastReader(mob_blast_results, logging).df
    if len(mob_blast_df) > 0:
        mob_blast_df = fixStart(mob_blast_df.drop(0).sort_values(['sseqid', 'sstart', 'send', 'bitscore'],
                                                                 ascending=[True, True, True, False]))
        mob_blast_df = remove_split_hits(mob_blast_df, 'qseqid', 'sseqid', 'sstart', 'send', 'bitscore').sort_values(
            ['sseqid', 'sstart', 'send', 'bitscore'], ascending=[True, True, True, False])
        mob_blast_df = recursive_filter_overlap_records(mob_blast_df, 5, 'sseqid', 'sstart', 'send',
                                                        'bitscore')

        add_biomarker_results(biomarker_df=mob_blast_df, df_column_name_biomarker='qseqid',
                              df_column_name_seqid='sseqid', contig_info=contig_info,
                              contig_info_type_key='relaxase_type(s)', contig_info_acs_key='relaxase_type_accession(s)',
                              delimeter='|')

    del (mob_blast_df)

    # blast mpf database
    logging.info("Blasting MPF sequences {} against {}".format(mpf_ref, fixed_fasta))
    tblastn(input_fasta=mpf_ref, blastdb=fixed_fasta, min_ident=min_mpf_ident, min_covs=min_mpf_cov,
            evalue=min_mpf_evalue, out_dir=tmp_dir,
            blast_results_file=mpf_blast_results, num_threads=num_threads, logging=logging)

    mpf_blast_df = BlastReader(mpf_blast_results, logging).df

    if len(mpf_blast_df) > 0:
        mpf_blast_df = fixStart(mpf_blast_df.drop(0)).sort_values(['sseqid', 'sstart', 'send', 'bitscore'],
                                                                  ascending=[True, True, True, False])
        mpf_blast_df = remove_split_hits(mpf_blast_df, 'qseqid', 'sseqid', 'sstart', 'send', 'bitscore').sort_values(
            ['sseqid', 'sstart', 'send', 'bitscore'], ascending=[True, True, True, False])
        mpf_blast_df = recursive_filter_overlap_records(mpf_blast_df, 5, 'sseqid', 'sstart', 'send',
                                                        'bitscore')

        logging.info("Filtering MPF blast results {} ".format(mpf_blast_results))

        add_biomarker_results(biomarker_df=mpf_blast_df, df_column_name_biomarker='qseqid',
                              df_column_name_seqid='sseqid', contig_info=contig_info,
                              contig_info_type_key='mpf_type', contig_info_acs_key='mpf_type_accession(s)',
                              delimeter='|')
    del (mpf_blast_results)

    # blast orit database
    logging.info("Blasting orit sequences {} against {}".format(plasmid_orit, fixed_fasta))
    blastn(input_fasta=plasmid_orit, blastdb=fixed_fasta, min_ident=min_rep_ident, min_cov=min_rep_cov,
           evalue=min_rep_evalue, min_length=80, out_dir=tmp_dir,
           blast_results_file=orit_blast_results, num_threads=num_threads, logging=logging)

    logging.info("Filtering orit blast results {} ".format(orit_blast_results))

    orit_blast_df = BlastReader(orit_blast_results, logging).df
    if len(orit_blast_df) > 0:
        orit_blast_df = recursive_filter_overlap_records(fixStart(
            orit_blast_df.drop(0).sort_values(['sseqid', 'sstart', 'send', 'bitscore'],
                                              ascending=[True, True, True, False])), 5, 'sseqid', 'sstart', 'send',
                                                         'bitscore')
        add_biomarker_results(biomarker_df=orit_blast_df, df_column_name_biomarker='qseqid',
                              df_column_name_seqid='sseqid', contig_info=contig_info,
                              contig_info_type_key='orit_type(s)', contig_info_acs_key='orit_accession(s)',
                              delimeter='|')

    del (orit_blast_df)

    # blast repetitive database
    if repetitive_mask_file is not None:
        logging.info("Blasting contigs against repetitive sequences db: {}".format(repetitive_mask_file))
        blastn(input_fasta=fixed_fasta, blastdb=repetitive_mask_file, min_ident=min_rpp_ident, min_cov=min_rpp_cov,
               evalue=min_rpp_evalue, min_length=min_length, out_dir=tmp_dir,
               blast_results_file=repetitive_blast_results, num_threads=num_threads, logging=logging)
        logging.info("Filtering repetitive blast results {} ".format(repetitive_blast_results))

        repetitive_blast_df = BlastReader(repetitive_blast_results, logging).df
        if len(repetitive_blast_df) > 0:
            repetitive_blast_df = recursive_filter_overlap_records(fixStart(
                repetitive_blast_df.drop(0).sort_values(['sseqid', 'sstart', 'send', 'bitscore'],
                                                        ascending=[True, True, True, False])), 5, 'qseqid', 'qstart',
                                                                   'qend',
                                                                   'bitscore')

            repetitive_list = repetitive_blast_df['qseqid'].tolist()

            # add filtering flag to contigs which are primarially a repetitive element
            for contig_id in repetitive_list:
                if contig_id in contig_info:
                    logging.info('Filtering contig: {} due to repetitive sequence'.format(contig_id))
                    contig_info[contig_id]['filtering_reason'] = 'repetitve element'
                else:
                    logging.error('Contig: {} not found in contig_df this is likely an error'.format(contig_id))

            add_biomarker_results(biomarker_df=repetitive_blast_df, df_column_name_biomarker='sseqid',
                                  df_column_name_seqid='qseqid', contig_info=contig_info,
                                  contig_info_type_key='repetitive_dna_type', contig_info_acs_key='repetitive_dna_id',
                                  delimeter='|', type_col_num=2, type_acs_num=1)

        del (repetitive_blast_df)
    return contig_info

def blast_mge(contig_fasta, mge_fasta,tmp_dir, min_length, logging, min_rpp_ident, min_rpp_cov, min_rpp_evalue,num_threads=1):
    # blast repetitive database
    blast_results = os.path.join(tmp_dir,"mge.blast.results.txt")
    blastn(input_fasta=mge_fasta, blastdb=contig_fasta, min_ident=min_rpp_ident, min_cov=min_rpp_cov,
           evalue=min_rpp_evalue, min_length=min_length, out_dir=tmp_dir,
           blast_results_file=blast_results, num_threads=num_threads, logging=logging)

    #return empty dataframe if no blast results generated
    if not os.path.isfile(blast_results) or os.path.getsize(blast_results) == 0:
        return {}

    df = BlastReader(blast_results, logging).df
    reduced_df = recursive_filter_overlap_records(fixStart(
        df.drop(0).sort_values(['sseqid', 'sstart', 'send', 'bitscore'],
                                                ascending=[True, True, True, False])), 5, 'sseqid', 'sstart',
        'send',
        'bitscore')

    results = {}
    for index,row in reduced_df.iterrows():
        contig_id = row['sseqid']
        if not contig_id in results:
            results[contig_id] = []
        results[contig_id].append(row.to_dict())

    return results

def writeMGEresults(contig_membership,mge_results,outfile):
    out_string = ["\t".join(MGE_INFO_HEADER)]
    if len(mge_results) == 0:
        return
    for contig_id in contig_membership['chromosome']:
        if not contig_id in mge_results:
            continue

        for i in range(0, len(mge_results[contig_id])):
            row = {}
            for field in MGE_INFO_HEADER:
                row[field] = ''
                if field in mge_results[contig_id]:
                    row[field] = mge_results[contig_id][field]

            id = mge_results[contig_id][i]['qseqid'].split('|')
            row['mge_id'] = id[0]
            row['mge_acs'] = id[1]
            row['mge_type'] = id[2]
            row['mge_subtype'] = id[3]
            row['mge_length'] = mge_results[contig_id][i]['qlen']
            row['mge_start'] = mge_results[contig_id][i]['qstart']
            row['mge_end'] = mge_results[contig_id][i]['qend']
            row['contig_start'] = mge_results[contig_id][i]['sstart']
            row['contig_end'] = mge_results[contig_id][i]['send']

            for field in contig_membership['chromosome'][contig_id]:
                if field in row:
                    row[field] = contig_membership['chromosome'][contig_id][field]

            out_string.append("\t".join([str(x) for x in list(row.values())]))

    for mobcluster in contig_membership['plasmid']:
        for contig_id in contig_membership['plasmid'][mobcluster]:
            if not contig_id in mge_results:
                continue

            for i in range(0, len(mge_results[contig_id])):
                row = {}
                for field in MGE_INFO_HEADER:
                    row[field] = ''
                    if field in mge_results[contig_id][i]:
                        row[field] = mge_results[contig_id][i][field]
                id = mge_results[contig_id][i]['qseqid'].split('|')
                row['mge_id'] = id[0]
                row['mge_acs'] = id[1]
                row['mge_type'] = id[2]
                row['mge_subtype'] = id[3]
                row['mge_length'] =  mge_results[contig_id][i]['qlen']
                row['mge_start'] = mge_results[contig_id][i]['qstart']
                row['mge_end'] = mge_results[contig_id][i]['qend']
                row['contig_start'] = mge_results[contig_id][i]['sstart']
                row['contig_end'] = mge_results[contig_id][i]['send']

                for field in contig_membership['plasmid'][mobcluster][contig_id]:
                    if field in row:
                        row[field] = contig_membership['plasmid'][mobcluster][contig_id][field]

                out_string.append("\t".join([str(x) for x in list(row.values())]))


    fh = open(outfile,'w')
    fh.write("\n".join(out_string))
    fh.close()


def create_biomarker_dataframe(parameters,id_mapping,logging):
    logging.info("Creating plasmid biomarkers report ...")
    data_frames = []
    for label in parameters:
        file = parameters[label]['file']  
        if not os.path.isfile(file):
            continue
        blast_df = pd.read_csv(file, header=0, sep="\t")
     
        if len(blast_df) == 0:
            continue
        blast_df['sseqid'] = blast_df['sseqid'].replace(id_mapping)
        blast_df['length'].astype('int64')
        blast_df['qcovs'].astype('float64')
        blast_df['qcovhsp'].astype('float64')
        blast_df['evalue'].astype('float64')
        blast_df['pident'].astype('float64')
        blast_df = blast_df.loc[blast_df['length'] >= parameters[label]['min_length']]
        blast_df = blast_df.loc[blast_df['qcovs'] >= parameters[label]['min_cov']]
        blast_df = blast_df.loc[blast_df['qcovhsp'] >= parameters[label]['min_hsp_cov']]
        blast_df = blast_df.loc[blast_df['evalue'] <= parameters[label]['evalue']]
        blast_df = blast_df.loc[blast_df['pident'] >= parameters[label]['min_ident']]
        blast_df = blast_df.reset_index(drop=True)

        blast_df = fixStart(blast_df)
        if label == 'replicon' or label == 'relaxase' or label == 'mate-pair-formation':
            blast_df = remove_split_hits(blast_df,'qseqid', 'sseqid', 'sstart', 'send', 'bitscore')
        blast_df.sort_values(['sseqid', 'sstart', 'send', 'bitscore'],ascending=[True, True, True, False], inplace=True)
    
        blast_df = recursive_filter_overlap_records(blast_df, 5, 'sseqid', 'sstart', 'send', 'bitscore')
        blast_df = blast_df.reset_index(drop=True)
        blast_df['biomarker'] = label
        data_frames.append(blast_df)
    if len(data_frames) > 0:    
        return pd.concat(data_frames)
    else:
        return pd.DataFrame()



import mob_suite.mob_typer
import os,sys
import pandas
import logging

TEST_ROOT = os.path.dirname(__file__)
LOG_FORMAT = '%(asctime)s %(levelname)s: %(message)s [in %(pathname)s:%(lineno)d]'
logging.basicConfig(format=LOG_FORMAT)
logger=logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def test_mean_and_multireplicon_frame():
    logger.info("Testing MOB-typer")
    if os.path.exists("run_test") == False:
        os.mkdir("run_test")
    args = [
        "--infile", os.path.dirname(__file__) + "/TestData/pCAV1453-208.fasta",
        "--out_file", TEST_ROOT + "/run_test/mobtyper_pCAV1453-208_results.txt",
    ]

    sys.argv[1:] = args
    mob_suite.mob_typer.main()
    assert os.path.exists(TEST_ROOT + "/run_test/mobtyper_pCAV1453-208_results.txt"), "Missing mobtyper_results.txt. Error MOB-typer did not run successfully"



#!/usr/bin/env python3

import logging
import os, re, shutil, sys, tempfile
from argparse import (ArgumentParser, ArgumentDefaultsHelpFormatter, RawDescriptionHelpFormatter)
from mob_suite.version import __version__
import mob_suite.mob_init
from collections import OrderedDict
from operator import itemgetter
from mob_suite.blast import BlastRunner
from mob_suite.wrappers import mash
from Bio import SeqIO

from mob_suite.utils import fix_fasta_header, \
    calcFastaStats, \
    verify_init, \
    check_dependencies, \
    read_sequence_info, \
    writeReport, \
    sort_biomarkers, \
    calc_md5, \
    gc_fraction, \
    read_fasta_dict, \
    identify_biomarkers, \
    parseMash, \
    determine_mpf_type, \
    hostrange, \
    dict_from_alt_key_list, \
    read_file_to_dict, \
    blast_mge, \
    writeMGEresults, \
    create_biomarker_dataframe

from mob_suite.constants import MOB_TYPER_REPORT_HEADER, \
    MOB_CLUSTER_INFO_HEADER, \
    default_database_dir, \
    LIT_PLASMID_TAXONOMY_HEADER


def init_console_logger(lvl=2):
    root = logging.getLogger()

    LOG_FORMAT = '%(asctime)s %(name)s %(levelname)s: %(message)s [in %(pathname)s:%(lineno)d]'
    logging_levels = [logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG]

    report_lvl = logging_levels[lvl]
    root.setLevel(report_lvl)  # set root logger level

    logging.basicConfig(format=LOG_FORMAT, level=report_lvl)
    return logging.getLogger(__name__)


def parse_args():
    "Parse the input arguments, use '-h' for help"

    class CustomFormatter(ArgumentDefaultsHelpFormatter, RawDescriptionHelpFormatter):
        pass

    parser = ArgumentParser(
        description="MOB-Typer: Plasmid typing and mobility prediction: {}".format(
            __version__), formatter_class=CustomFormatter)
    parser.add_argument('-i', '--infile', type=str, required=True, help='Input assembly fasta file to process')
    parser.add_argument('-o', '--out_file', type=str, required=True, help='Output file to write results')
    parser.add_argument('--biomarker_report_file', type=str, required=False, help='Output file for biomarker blast results')
    parser.add_argument('-g', '--mge_report_file', type=str, required=False, help='Output file for MGE results')
    parser.add_argument('-a', '--analysis_dir', type=str, required=False,
                        help='Working directory for storing temporary results')
    parser.add_argument('-n', '--num_threads', type=int, required=False, help='Number of threads to be used', default=1)
    parser.add_argument('-s', '--sample_id', type=str, required=False, help='Sample Prefix for reports')
    parser.add_argument('-x', '--multi', required=False, help='Treat each sequence as an independant plasmid',
                        action='store_true')
    parser.add_argument('--min_rep_evalue', type=float, required=False,
                        help='Minimum evalue threshold for replicon blastn',
                        default=0.00001)
    parser.add_argument('--min_mob_evalue', type=float, required=False,
                        help='Minimum evalue threshold for relaxase tblastn',
                        default=0.00001)
    parser.add_argument('--min_con_evalue', type=float, required=False, help='Minimum evalue threshold for contig blastn',
                        default=0.00001)

    parser.add_argument('--min_length', type=str, required=False, help='Minimum length of blast hits',
                        default=500)
    parser.add_argument('--min_rep_ident', type=int, required=False, help='Minimum sequence identity for replicons',
                        default=80)
    parser.add_argument('--min_mob_ident', type=int, required=False, help='Minimum sequence identity for relaxases',
                        default=80)
    parser.add_argument('--min_con_ident', type=int, required=False, help='Minimum sequence identity for contigs',
                        default=80)
    parser.add_argument('--min_rpp_ident', type=int, required=False,
                        help='Minimum sequence identity for MGE', default=80)

    parser.add_argument('--min_rep_cov', type=int, required=False,
                        help='Minimum percentage coverage of replicon query by input assembly',
                        default=80)

    parser.add_argument('--min_mob_cov', type=int, required=False,
                        help='Minimum percentage coverage of relaxase query by input assembly',
                        default=80)

    parser.add_argument('--min_con_cov', type=int, required=False,
                        help='Minimum percentage coverage of assembly contig by the plasmid reference database to be considered',
                        default=70)
    parser.add_argument('--min_rpp_cov', type=int, required=False,
                        help='Minimum percentage coverage of MGE',
                        default=80)
    parser.add_argument('--min_rpp_evalue', type=float, required=False,
                        help='Minimum evalue threshold for repetitve elements blastn',
                        default=0.00001)

    parser.add_argument('--min_overlap', type=int, required=False,
                        help='Minimum overlap of fragments',
                        default=10)

    parser.add_argument('-k', '--keep_tmp', required=False, help='Do not delete temporary file directory',
                        action='store_true')

    parser.add_argument('--debug', required=False, help='Show debug information', action='store_true')

    parser.add_argument('--plasmid_mash_db', type=str, required=False,
                        help='Companion Mash database of reference database',
                        default=os.path.join(default_database_dir,
                                             'ncbi_plasmid_full_seqs.fas.msh'))
    parser.add_argument('-m', '--plasmid_meta', type=str, required=False,
                        help='MOB-cluster plasmid cluster formatted file matched to the reference plasmid db',
                        default=os.path.join(default_database_dir,
                                             'clusters.txt'))
    parser.add_argument('--plasmid_db_type', type=str, required=False, help='Blast database type of reference database',
                        default='blastn')
    parser.add_argument('--plasmid_replicons', type=str, required=False, help='Fasta of plasmid replicons',
                        default=os.path.join(default_database_dir,
                                             'rep.dna.fas'))
    parser.add_argument('--repetitive_mask', type=str, required=False, help='Fasta of known repetitive elements',
                        default=os.path.join(default_database_dir,
                                             'repetitive.dna.fas'))
    parser.add_argument('--plasmid_mob', type=str, required=False, help='Fasta of plasmid relaxases',
                        default=os.path.join(default_database_dir,
                                             'mob.proteins.faa'))
    parser.add_argument('--plasmid_mpf', type=str, required=False, help='Fasta of known plasmid mate-pair proteins',
                        default=os.path.join(default_database_dir,
                                             'mpf.proteins.faa'))
    parser.add_argument('--plasmid_orit', type=str, required=False, help='Fasta of known plasmid oriT dna sequences',
                        default=os.path.join(default_database_dir,
                                             'orit.fas'))
    parser.add_argument('-d', '--database_directory',
                        default=default_database_dir,
                        required=False,
                        help='Directory you want to use for your databases. If the databases are not already '
                             'downloaded, they will be downloaded automatically. Defaults to {}'.format(
                            default_database_dir))
    parser.add_argument('--primary_cluster_dist', type=float, required=False,
                        help='Mash distance for assigning primary cluster id 0 - 1', default=0.06)
    parser.add_argument('-V', '--version', action='version', version="%(prog)s " + __version__)

    return parser.parse_args()


def initMOBTyperReportTemplate(header):
    data = {}
    for i in header:
        data[i] = '-'
    return data


def main():
    args = parse_args()

    if args.debug:
        logger = init_console_logger(3)
    else:
        logger = init_console_logger(2)

    logger.info('Running Mob-typer version {}'.format(__version__))

    logger.info('Processing fasta file {}'.format(args.infile))

    if not os.path.isfile(args.infile):
        logger.info('Error, fasta file does not exist {}'.format(args.infile))
        sys.exit()

    if not args.analysis_dir:
        tmp_dir = tempfile.TemporaryDirectory(dir=tempfile.gettempdir()).name
    else:
        tmp_dir = args.analysis_dir


    if not os.path.isdir(tmp_dir):
        logger.info('Creating analysis directory {}'.format(tmp_dir))
        os.makedirs(tmp_dir, 0o755)


    if not isinstance(args.num_threads, int):
        logger.info('Error number of threads must be an integer, you specified "{}"'.format(args.num_threads))

    database_dir = os.path.abspath(args.database_directory)

    if args.sample_id is None:
        sample_id = re.sub(r"\.(fasta|fa|fas){1,1}", "", os.path.basename(args.infile))
    else:
        sample_id = args.sample_id

    # Script arguments
    input_fasta = args.infile
    report_file = args.out_file
    mge_report_file = args.mge_report_file
    num_threads = int(args.num_threads)
    keep_tmp = args.keep_tmp
    biomarker_report_file = args.biomarker_report_file

    if args.multi:
        multi = True
    else:
        multi = False

    if not (args.primary_cluster_dist >= 0 and args.primary_cluster_dist <= 1):
        logging.error('Error distance thresholds must be between 0 - 1: {}'.format(args.primary_cluster_dist))
        sys.exit()
    else:
        primary_distance = float(args.primary_cluster_dist)

    min_length = int(args.min_length)
    ETE3DBTAXAFILE = os.path.abspath(database_dir + "/taxa.sqlite")

    if database_dir == default_database_dir:
        mob_ref = args.plasmid_mob
        mash_db = args.plasmid_mash_db
        replicon_ref = args.plasmid_replicons
        plasmid_meta = args.plasmid_meta
        mpf_ref = args.plasmid_mpf
        plasmid_orit = args.plasmid_orit
        repetitive_mask_file = args.repetitive_mask
        verify_init(logger, database_dir)
    else:
        mob_ref = os.path.join(database_dir, 'mob.proteins.faa')
        mash_db = os.path.join(database_dir, 'ncbi_plasmid_full_seqs.fas.msh')
        replicon_ref = os.path.join(database_dir, 'rep.dna.fas')
        plasmid_meta = os.path.join(database_dir, 'clusters.txt')
        mpf_ref = os.path.join(database_dir, 'mpf.proteins.faa')
        plasmid_orit = os.path.join(database_dir, 'orit.fas')
        repetitive_mask_file = os.path.join(database_dir, 'repetitive.dna.fas')


    LIT_PLASMID_TAXONOMY_FILE = os.path.join(database_dir, "host_range_literature_plasmidDB.txt")
    NCBI_PLASMID_TAXONOMY_FILE = plasmid_meta

    fixed_fasta = os.path.join(tmp_dir, 'fixed.input.fasta')
    replicon_blast_results = os.path.join(tmp_dir, 'replicon_blast_results.txt')
    mob_blast_results = os.path.join(tmp_dir, 'mobtyper_blast_results.txt')
    mpf_blast_results = os.path.join(tmp_dir, 'mpf_blast_results.txt')
    orit_blast_results = os.path.join(tmp_dir, 'orit_blast_results.txt')
    repetitive_blast_results = os.path.join(tmp_dir, 'repetitive_blast_results.txt')

    if os.path.isfile(mob_blast_results):
        os.remove(mob_blast_results)
    if os.path.isfile(mpf_blast_results):
        os.remove(mpf_blast_results)
    if os.path.isfile(orit_blast_results):
        os.remove(orit_blast_results)
    if os.path.isfile(replicon_blast_results):
        os.remove(replicon_blast_results)

    # Input numeric params

    min_rep_ident = float(args.min_rep_ident)
    min_mob_ident = float(args.min_mob_ident)
    min_ori_ident = float(args.min_rep_ident)
    min_mpf_ident = float(args.min_mob_ident)
    min_rpp_ident = float(args.min_rpp_ident)

    idents = {'min_rep_ident': min_rep_ident, 'min_mob_ident': min_mob_ident, 'min_ori_ident': min_ori_ident}

    for param in idents:

        value = float(idents[param])

        if value < 60:
            logger.error("Error: {} is too low, please specify an integer between 70 - 100".format(param))
            sys.exit(-1)
        if value > 100:
            logger.error("Error: {} is too high, please specify an integer between 70 - 100".format(param))
            sys.exit(-1)

    min_rep_cov = float(args.min_rep_cov)
    min_mob_cov = float(args.min_mob_cov)
    min_ori_cov = float(args.min_rep_cov)
    min_mpf_cov = float(args.min_mob_cov)
    min_rpp_cov = float(args.min_rpp_cov)

    covs = {'min_rep_cov': min_rep_cov, 'min_mob_cov': min_mob_cov, 'min_con_cov': min_ori_cov,
            'min_rpp_cov': min_ori_cov}

    for param in covs:

        value = float(covs[param])

        if value < 50:
            logger.error("Error: {} is too low, please specify an integer between 50 - 100".format(param))
            sys.exit(-1)
        if value > 100:
            logger.error("Error: {} is too high, please specify an integer between 50 - 100".format(param))
            sys.exit(-1)

    min_rep_evalue = float(args.min_rep_evalue)
    min_mob_evalue = float(args.min_mob_evalue)
    min_ori_evalue = float(args.min_rep_evalue)
    min_mpf_evalue = float(args.min_mob_evalue)
    min_rpp_evalue = float(args.min_rpp_evalue)

    evalues = {'min_rep_evalue': min_rep_evalue, 'min_mob_evalue': min_mob_evalue, 'min_con_evalue': min_ori_evalue}

    for param in evalues:

        value = float(evalues[param])

        if value > 1:
            logger.error("Error: {} is too high, please specify an float evalue between 0 to 1".format(param))
            sys.exit(-1)

    check_dependencies(logger)

    needed_dbs = [replicon_ref, mob_ref, mash_db, mpf_ref]

    for db in needed_dbs:
        if (not os.path.isfile(db)):
            logger.info('Warning! Needed database missing "{}"'.format(db))
            mob_suite.mob_init.main()

    if not os.path.isdir(tmp_dir):
        os.mkdir(tmp_dir, 0o755)

    # Get cluster information
    reference_sequence_meta = read_sequence_info(plasmid_meta, MOB_CLUSTER_INFO_HEADER)

    # initilize master record tracking
    id_mapping = fix_fasta_header(input_fasta, fixed_fasta)
    contig_info = {}
    with open(fixed_fasta, "r") as handle:
        for record in SeqIO.parse(handle, "fasta"):
            id = str(record.id)
            contig_info[id] = {}
            for feature in MOB_TYPER_REPORT_HEADER:
                contig_info[id][feature] = ''
            seq = str(record.seq)
            contig_info[id]['md5'] = calc_md5(seq)
            contig_info[id]['gc'] = gc_fraction(seq)
            contig_info[id]['size'] = len(seq)
            contig_info[id]['contig_id'] = id
            contig_info[id]['sample_id'] = sample_id
    handle.close()

    # Makeblastdb
    blast_runner = BlastRunner(fixed_fasta, tmp_dir)
    build_success = blast_runner.makeblastdb(fixed_fasta, 'nucl', logging=logging)
    if build_success == False:
        logging.error("Could not build blast database, check error messages..cannot continue")
        sys.exit()

    # run individual marker blasts

    contig_info = identify_biomarkers(contig_info, fixed_fasta, tmp_dir, 25, logging, \
                                      replicon_ref, min_rep_ident, min_rep_cov, min_rep_evalue, replicon_blast_results, \
                                      mob_ref, min_mob_ident, min_mob_cov, min_mob_evalue, mob_blast_results, \
                                      mpf_ref, min_mpf_ident, min_mpf_cov, min_mpf_evalue, mpf_blast_results, \
                                      None, None, None, None, \
                                      plasmid_orit, orit_blast_results, repetitive_blast_results, \
                                      num_threads=num_threads)

    m = mash()
    mobtyper_results = []

    mash_input_fasta = fixed_fasta + '.msh'

    ncbi = dict_from_alt_key_list(
        read_file_to_dict(NCBI_PLASMID_TAXONOMY_FILE, MOB_CLUSTER_INFO_HEADER, separater="\t"),
        "sample_id")
    lit = dict_from_alt_key_list(
        read_file_to_dict(LIT_PLASMID_TAXONOMY_FILE, LIT_PLASMID_TAXONOMY_HEADER, separater="\t"), "sample_id")

    if multi:
        m.mashsketch(input_fasta=fixed_fasta, output_path=mash_input_fasta, sketch_ind=True, num_threads=num_threads)
        mash_results = parseMash(
            m.run_mash(reference_db=mash_db, input_fasta=mash_input_fasta, table=False, num_threads=num_threads))

        for seq_id in mash_results:
            record = {}
            for field in MOB_TYPER_REPORT_HEADER:
                if field in contig_info[seq_id]:
                    record[field] = contig_info[seq_id][field]
                else:
                    record[field] = ''
            record['sample_id'] = seq_id
            record['num_contigs'] = 1
            distances = OrderedDict(sorted(mash_results[seq_id].items(), key=itemgetter(1), reverse=False))

            for mash_neighbor_id in distances:
                dist = distances[mash_neighbor_id]
                if mash_neighbor_id not in reference_sequence_meta:
                    continue
                else:
                    record['mash_nearest_neighbor'] = mash_neighbor_id
                    record['mash_neighbor_distance'] = dist
                    record['primary_cluster_id'] = reference_sequence_meta[mash_neighbor_id]['primary_cluster_id']
                    record['secondary_cluster_id'] = reference_sequence_meta[mash_neighbor_id]['secondary_cluster_id']
                    record['mash_neighbor_identification'] = reference_sequence_meta[mash_neighbor_id]['organism']
                    break
            mobtyper_results.append(record)

    else:
        m.mashsketch(input_fasta=fixed_fasta, output_path=mash_input_fasta, sketch_ind=False, num_threads=num_threads)
        mash_results = parseMash(
            m.run_mash(reference_db=mash_db, input_fasta=mash_input_fasta, table=False, num_threads=num_threads))
        record = {}

        for field in MOB_TYPER_REPORT_HEADER:
            record[field] = ''

        record['sample_id'] = sample_id
        fastaSeqStats = calcFastaStats(fixed_fasta)
        record['md5'] = fastaSeqStats['md5']
        record['size'] = fastaSeqStats['size']
        record['num_contigs'] = fastaSeqStats['num_seq']
        record['gc'] = fastaSeqStats['gc_content']
        record['mash_nearest_neighbor'] = '-'
        record['mash_neighbor_distance'] = 1
        record['primary_cluster_id'] = '-'
        record['secondary_cluster_id'] = '-'
        record['mash_neighbor_identification'] = '-'

        for seq_id in mash_results:
            distances = OrderedDict(sorted(mash_results[seq_id].items(), key=itemgetter(1), reverse=False))
            mash_neighbor_id = next(iter(distances))
            dist = distances[mash_neighbor_id]
            if mash_neighbor_id not in reference_sequence_meta:
                continue
            record['mash_nearest_neighbor'] = mash_neighbor_id
            record['mash_neighbor_distance'] = dist
            record['primary_cluster_id'] = reference_sequence_meta[mash_neighbor_id]['primary_cluster_id']
            record['secondary_cluster_id'] = reference_sequence_meta[mash_neighbor_id]['secondary_cluster_id']
            record['mash_neighbor_identification'] = reference_sequence_meta[mash_neighbor_id]['organism']

        record['rep_type(s)'] = []
        record['rep_type_accession(s)'] = []
        record['relaxase_type(s)'] = []
        record['relaxase_type_accession(s)'] = []
        record['mpf_type'] = []
        record['mpf_type_accession(s)'] = []
        record['orit_type(s)'] = []
        record['orit_accession(s)'] = []

        for seq_id in contig_info:
            record['rep_type(s)'].append(contig_info[seq_id]['rep_type(s)'])
            record['rep_type_accession(s)'].append(contig_info[seq_id]['rep_type_accession(s)'])
            record['relaxase_type(s)'].append(contig_info[seq_id]['relaxase_type(s)'])
            record['relaxase_type_accession(s)'].append(contig_info[seq_id]['relaxase_type_accession(s)'])
            record['mpf_type'].append(contig_info[seq_id]['mpf_type'])
            record['mpf_type_accession(s)'].append(contig_info[seq_id]['mpf_type_accession(s)'])
            record['orit_type(s)'].append(contig_info[seq_id]['orit_type(s)'])
            record['orit_accession(s)'].append(contig_info[seq_id]['orit_accession(s)'])

        for field in record:
            tmp = []
            if record[field] == None:
                continue
            if isinstance(record[field], list):
                length = len(record[field])
                for i in range(0, length):
                    tmp += record[field][i].split(',')
            elif isinstance(record[field], str) and len(record[field]) > 0:
                tmp += record[field].split(',')
            if len(tmp) > 0:
                record[field] = []
                for d in tmp:
                    if len(d) > 0:
                        record[field].append(d)

        mobtyper_results.append(record)



    for i in range(0, len(mobtyper_results)):
        record = mobtyper_results[i]
        sample_id = record['sample_id']
        if isinstance(record['sample_id'], list):
            sample_id = record['sample_id'][0]
        if sample_id in id_mapping:
            original_id = id_mapping[sample_id]
            record['sample_id'] = original_id
        bio_markers = sort_biomarkers({0: {'types': record['rep_type(s)'], 'acs': record['rep_type_accession(s)']},
                                       1: {'types': record['relaxase_type(s)'],
                                           'acs': record['relaxase_type_accession(s)']},
                                       2: {'types': record['mpf_type'], 'acs': record['mpf_type_accession(s)']},
                                       3: {'types': record['orit_type(s)'], 'acs': record['orit_accession(s)']}, })

        record['rep_type(s)'] = bio_markers[0]['types']
        record['rep_type_accession(s)'] = bio_markers[0]['acs']
        record['relaxase_type(s)'] = bio_markers[1]['types']
        record['relaxase_type_accession(s)'] = bio_markers[1]['acs']
        record['mpf_type'] = bio_markers[2]['types']
        record['mpf_type_accession(s)'] = bio_markers[2]['acs']
        record['orit_type(s)'] = bio_markers[3]['types']
        record['orit_accession(s)'] = bio_markers[3]['acs']

        if (isinstance(record['mash_neighbor_distance'], float) or isinstance(record['mash_neighbor_distance'],
                                                                              int)) and record[
            'mash_neighbor_distance'] <= primary_distance:
            mob_cluster_id = record['primary_cluster_id']
        else:
            mob_cluster_id = None

        # Patches that sometimes results are concatonated into strings if contigs are merged into a single results
        if isinstance(record['rep_type(s)'], list):
            record['rep_type(s)'] = ",".join(record['rep_type(s)'])
        if isinstance(record['relaxase_type_accession(s)'], list):
            record['relaxase_type_accession(s)'] = ",".join(record['relaxase_type_accession(s)'])

        host_range = hostrange(record['rep_type(s)'].split(','), record['relaxase_type_accession(s)'].split(','),
                               mob_cluster_id, ncbi, lit, ETE3DBTAXAFILE, database_dir)

        for field in host_range:
            record[field] = host_range[field]

        if isinstance(record['mpf_type'], list):
            record['mpf_type'] = determine_mpf_type(record['mpf_type'])
        elif isinstance(record['mpf_type'], str):
            record['mpf_type'] = determine_mpf_type(record['mpf_type'].split(','))

        for field in record:
            if isinstance(record[field], list):
                record[field] = ",".join(record[field])

        record['predicted_mobility'] = 'non-mobilizable'
        if len(record['relaxase_type(s)']) > 0 and len(record['mpf_type']):
            record['predicted_mobility'] = 'conjugative'
        elif len(record['relaxase_type(s)']) > 0 or len(record['orit_type(s)']) > 0:
            record['predicted_mobility'] = 'mobilizable'

        mobtyper_results[i] = record

    writeReport(mobtyper_results, MOB_TYPER_REPORT_HEADER, report_file)

    # Peform MGE detection
    if mge_report_file is not None:
        mge_results = blast_mge(fixed_fasta, repetitive_mask_file, tmp_dir, min_length,
                                logging, min_rpp_ident, min_rpp_cov, min_rpp_evalue, num_threads)

        tmp = {}
        for contig_id in mge_results:
            if contig_id in id_mapping:
                label = id_mapping[contig_id]
            else:
                continue
            tmp[label] = mge_results[contig_id]

        mge_results = tmp
        contig_memberships = {'chromosome': {}, 'plasmid': {}}
        for i in range(0, len(mobtyper_results)):
            primary_cluster_id = mobtyper_results[i]['primary_cluster_id']
            if not primary_cluster_id in contig_memberships['plasmid']:
                contig_memberships['plasmid'][primary_cluster_id] = {}
            contig_id = mobtyper_results[i]['sample_id']
            mobtyper_results[i]['molecule_type'] = 'plasmid'
            mobtyper_results[i]['contig_id'] = contig_id
            contig_memberships['plasmid'][primary_cluster_id][contig_id] = mobtyper_results[i]

        if len(mge_results) > 0:
            writeMGEresults(contig_memberships, mge_results, mge_report_file)
            logger.info("MOB-typer MGE results written to {}".format(mge_report_file))
        else:
            logger.info("No MOB-typer MGE to write")

    if biomarker_report_file is not None:
            
        biomarker_params = {
            'oriT': {
                'file':orit_blast_results,
                'min_length': 80,
                'min_cov':min_rep_cov,
                'min_hsp_cov': 15,
                'evalue':min_rep_evalue,
                'min_ident':min_rep_ident
            },
            'replicon': {
                'file':replicon_blast_results,
                'min_length': 80,
                'min_cov':min_rep_cov,
                'min_hsp_cov': 25,
                'evalue':min_rep_evalue,
                'min_ident':min_rep_ident
            },
            'relaxase': {
                'file':mob_blast_results,
                'min_length': 40,
                'min_cov':min_mob_cov,
                'min_hsp_cov': 25,
                'evalue':min_mob_evalue,
                'min_ident':min_mob_ident
            },        
            'mate-pair-formation': {
                'file':mpf_blast_results,
                'min_length': 40,
                'min_cov':min_mpf_cov,
                'min_hsp_cov': 25,
                'evalue':min_mpf_evalue,
                'min_ident':min_mpf_ident
            },

        }
        
        biomarker_df = create_biomarker_dataframe(biomarker_params,id_mapping,logging)
        biomarker_df.to_csv(biomarker_report_file,header=True,sep="\t",index=False)


    if not keep_tmp:
        shutil.rmtree(tmp_dir)
    logger.info("MOB-typer completed and results written to {}".format(report_file))


# call main function
if __name__ == '__main__':
    main()



import os, time, random, sys
from statistics import mean
from Bio import SeqIO
import pandas as pd
import logging
import itertools
import mob_suite.mob_typer



SAMPLES_WITH_ISSUES_DF = pd.DataFrame(columns=['plasmid_id','issue_type','marker_type','marker_name','marker_accession'])
TEST_ROOT = os.path.dirname(__file__)
PACKAGE_ROOT = os.path.dirname(mob_suite.__file__)
LOG_FORMAT = '%(asctime)s %(levelname)s: %(message)s [in %(pathname)s:%(lineno)d]'
NRANDOMSAMPLES2TEST = 300
COMPLEX_CASES_LIST = ['CP011291', 'CP045773', 'KU932024', 'NC_002134', 'CP033122', 'CP031575', 'CP021102',
                      'CP033949', 'CP025336', 'CP025336', 'CP028546']

logging.basicConfig(format=LOG_FORMAT)
logger=logging.getLogger(__name__)
logger.setLevel(logging.INFO)

def clean_missing_dashes(L):
    return [i for i in L if i != '-']


def test_mean_and_multireplicon_frame(mode = None):
    if os.path.exists("run_test") == False:
        os.mkdir("run_test")

    if mode == None:
        return 0
    logger.info("Testing MOB-typer for biomarker and mobtyper reports consistency in terms of biomarker names (not accessions)")
    logger.info(f"Selected testing mode: '{mode}'")
    elapsed_time_list=list()
    args = [
        "--num_threads", "4",
        "--infile", TEST_ROOT + "/run_test/tmp_plasmid.fasta",
        "--out_file", TEST_ROOT + "/run_test/mobtyper_report.txt",
        "--biomarker_report_file", TEST_ROOT + "/run_test/biomarker_report.txt"
    ]
    sys.argv[1:] = args
    
    with open(os.path.join(PACKAGE_ROOT + "/databases/ncbi_plasmid_full_seqs.fas")) as handle:
        records = list(SeqIO.parse(handle, "fasta"))
        if mode == "random sampling":
            records=random.sample(records, NRANDOMSAMPLES2TEST)
        n_records = len([rec.id for rec in records])
        for n, record in enumerate(records):
            if mode == "list of sample ids":
                if record.id not in COMPLEX_CASES_LIST:
                    continue
            start_time = time.time()
            logger.info(f"*** Sample #{n+1} out of {n_records} -> record ID {record.id} ***")
            SeqIO.write(record, TEST_ROOT + "/run_test/tmp_plasmid.fasta", "fasta")
            mob_suite.mob_typer.main()
            mobtyper_report_df = pd.read_csv(TEST_ROOT + "/run_test/mobtyper_report.txt",sep="\t")
            if os.path.getsize( TEST_ROOT + "/run_test/biomarker_report.txt") == 1:
                logger.warning(f'Empty biomarker report for {record.id}')
                continue

            # check biomarker report in sync with mobtyper report    
            mobtyper_biomarker_df = pd.read_csv(TEST_ROOT + "/run_test/biomarker_report.txt",sep="\t")   
            mobtyper_marker_types = clean_missing_dashes(list(itertools.chain(* mobtyper_report_df[['orit_type(s)','rep_type(s)', 'relaxase_type(s)', 'mpf_type']].stack().str.split(','))))
            mobtyper_marker_accessions=clean_missing_dashes(list(itertools.chain(* mobtyper_report_df[['rep_type_accession(s)','relaxase_type_accession(s)', 'mpf_type_accession(s)', 'orit_accession(s)']].stack().str.split(','))))
            
            #for accession, biomarker_name in mobtyper_biomarker_df.qseqid.str.split('|'):
            for index, row in mobtyper_biomarker_df.iterrows():
                accession, biomarker_name = row.qseqid.split('|')       
                match_result = mobtyper_report_df.apply(lambda item: item.astype(str).str.contains(accession) , axis=0).any().any(axis=None)
                if match_result == False:
                    logger.info(f"Issue with {record.id}. Marker with accession {biomarker_name}:{accession} is not found in mobtyper report")
                    SAMPLES_WITH_ISSUES_DF.loc[SAMPLES_WITH_ISSUES_DF.shape[0],:]=[
                        record.id,'missing in mobtyper report',row.biomarker, biomarker_name, accession]
                if biomarker_name in mobtyper_marker_types:
                    mobtyper_marker_types.remove(biomarker_name)
                if accession in mobtyper_marker_accessions:
                    mobtyper_marker_accessions.remove(accession)   
            if len(mobtyper_marker_types) != 0:
                logger.error(f'Following biomarker types are missing in biomarkers report:{",".join(mobtyper_marker_types)}')
                SAMPLES_WITH_ISSUES_DF.loc[SAMPLES_WITH_ISSUES_DF.shape[0],:]=[
                    record.id,'missing in biomarkers report',row.biomarker,",".join(mobtyper_marker_types),",".join(mobtyper_marker_accessions)]
        
            elapsed_time=time.time() - start_time
            elapsed_time_list.append(elapsed_time)
            logger.info("Elapsed {0:.0f} seconds for {1} sample typing".format(elapsed_time, record.id))
            logger.info("*** Estimated time to completion {0:.2f} hours (mean time per test {1:.0f} s) ***".format((n_records-n)*mean(elapsed_time_list)/3600, mean(elapsed_time_list) ))
           

    SAMPLES_WITH_ISSUES_DF.to_csv(TEST_ROOT+'/plasmid_accessions_with_issues.txt', sep="\t", index=False)          
    n_issues2solve = len(SAMPLES_WITH_ISSUES_DF['plasmid_id'].unique()) 
    logger.info(f"Analysis completed with {n_issues2solve} issues ...")
    if len(SAMPLES_WITH_ISSUES_DF) == 0:
        logger.info("PASSED ALL TEST, CONGRATULATIONS! 100% consistency!")
    else:
        logger.info("Oh well, more work to do with {0:.1f}% success rate".format((NRANDOMSAMPLES2TEST-n_issues2solve)/n_records*100))
        logger.info("*** Problematic samples listed in {} ***".format(TEST_ROOT+'/plasmid_accessions_with_issues.txt'))       


    assert len(SAMPLES_WITH_ISSUES_DF) == 0, "Test failed with {0:.1f}% success rate".format((NRANDOMSAMPLES2TEST-n_issues2solve)/n_records*100)


test_mean_and_multireplicon_frame(mode="list of sample ids")
test_mean_and_multireplicon_frame(mode="random sampling")


from subprocess import Popen, PIPE, check_call
import os, sys, logging
from mob_suite.blast import BlastRunner
from mob_suite.blast import BlastReader

class mash:
    def __init__(self):
        return

    def run_mash(self, reference_db, input_fasta, table=False, num_threads=1):

        if table:
            p = Popen(['mash', "dist", "-t", "-p", str(num_threads), reference_db, input_fasta],
                      stdout=PIPE,
                      stderr=PIPE)
            (stdout, stderr) = p.communicate()
        else:
            p = Popen(['mash', "dist", "-p", str(num_threads), reference_db, input_fasta],
                      stdout=PIPE,
                      stderr=PIPE)
            (stdout, stderr) = p.communicate()
        if len(str(stderr)) > 0 and str(stderr) != "b''":
            logging.info('{}'.format(stderr))
        return stdout

    def run_mash_screen(self, reference_db, input_fasta, winner_take_all=True, num_threads=1):

        if winner_take_all:
            p = Popen(['mash', "screen", "-p", str(num_threads), '-w', '-i','0', reference_db, input_fasta],
                      stdout=PIPE,
                      stderr=PIPE)
            (stdout, stderr) = p.communicate()
        else:
            p = Popen(['mash', "screen", "-p", str(num_threads), '-i','0', reference_db, input_fasta],
                      stdout=PIPE,
                      stderr=PIPE)
            (stdout, stderr) = p.communicate()

        logging.info(
            '{}'.format(stderr))
        return stdout


    def mashsketch(self, input_fasta, output_path, sketch_ind=True, num_threads=1, kmer_size=21, sketch_size=1000):
        if output_path == '':
            os.path.dirname(input_fasta)

        if sketch_ind:
            p = Popen(['mash', "sketch",
                       "-p", str(num_threads),
                       "-i",
                       "-o", output_path,
                       "-k", str(kmer_size),
                       "-s", str(sketch_size), input_fasta],
                      stdout=PIPE,
                      stderr=PIPE)
        else:
            p = Popen(['mash', "sketch",
                       "-p", str(num_threads),
                       "-o", output_path,
                       "-k", str(kmer_size),
                       "-s", str(sketch_size), input_fasta],
                      stdout=PIPE,
                      stderr=PIPE)
        p.wait()
        stdout = p.stdout.read()
        stderr = p.stderr.read()


class detectCircularity:
    ### Method adapted from Berokka https://github.com/tseemann/berokka by Torsten Seemann
    def __init__(self):
        return

    def run(self,input_fasta, output_path, logging, min_cov=1, min_ident=1, evalue=1, num_threads=1, min_length=25):
        blast_results_file = os.path.join(output_path, 'circularize.blast.txt')
        self.run_blast(input_fasta=input_fasta,
                       output_path=output_path,
                       blast_results_file=blast_results_file,
                       logging=logging,
                       min_cov=min_cov,
                       min_ident=min_ident,
                       evalue=evalue,
                       num_threads=num_threads,
                       min_length=min_length)
        return (self.overhangDetection(blast_results_file,logging))

    def run_blast(self,input_fasta,output_path,blast_results_file,logging,min_cov=1,min_ident=1,evalue=1,num_threads=1,min_length=25):
        blast_runner = BlastRunner(input_fasta, output_path)
        blast_runner.makeblastdb(input_fasta, 'nucl',logging)
        blast_runner.run_blast(query_fasta_path=input_fasta, blast_task='megablast', db_path=input_fasta,
                               db_type='nucl', min_cov=min_cov, min_ident=min_ident, evalue=evalue,
                               blast_outfile=blast_results_file, num_threads=num_threads, word_size=11,logging=logging)

        if os.path.getsize(blast_results_file) == 0:
            fh = open(blast_results_file, 'w', encoding="utf-8")
            fh.write('')
            fh.close()
            return dict()

        blast_df = BlastReader(blast_results_file,logging).df
        blast_df = blast_df.loc[blast_df['length'] >= min_length]
        blast_df = blast_df.reset_index(drop=True)
        blast_df.to_csv(blast_results_file, sep='\t', header=False, line_terminator='\n', index=False)

    def overhangDetection(self,blast_results_file,logging,min_length=25):
        if os.path.getsize(blast_results_file) == 0:
            return dict()

        blast_df = BlastReader(blast_results_file,logging).df.sort_values(['qseqid', 'qstart', 'qend', 'bitscore'], ascending=[True, True, True, False])

        circular_contigs = {}


        for index, row in blast_df.iterrows():
            contig_id_query = row['qseqid']
            contig_id_subject = row['sseqid']
            contig_start_subject = int(row['sstart'])
            contig_end_subject = int(row['send'])
            contig_start_query = int(row['qstart'])
            contig_end_query = int(row['qend'])
            contig_length = int(row['qlen'])
            length = int(row['length'])


            if contig_id_query != contig_id_subject and contig_id_subject != "ref|{}|".format(contig_id_query):
                continue

            if contig_start_query != 1 or length < min_length:

                continue

            if contig_start_query  == contig_start_subject and contig_end_query == contig_end_subject:

                continue



            if contig_start_query == 1  and contig_end_subject == contig_length:
                circular_contigs[contig_id_query] = 'Circular: Overlap {} bp'.format(length)
        return circular_contigs








from datetime import datetime
import logging, sys
from pathlib import Path

from subprocess import Popen, PIPE
import os

import pandas as pd
from pandas.errors import EmptyDataError



BLAST_TABLE_COLS = '''
qseqid
sseqid
qlen
slen
qstart
qend
sstart
send
length
mismatch
pident
qcovhsp
qcovs
sstrand
evalue
bitscore
'''.strip().split('\n')

#inherit logger
logger = logging.getLogger(__name__)
logger.setLevel(logging.getLogger().getEffectiveLevel())



class BlastRunner:

    def __init__(self, fasta_path, tmp_work_dir):
        self.fasta_path = fasta_path

    def makeblastdb(self,fasta_path,dbtype,logging,parse_seqids=False):
        if parse_seqids:
            p = Popen(['makeblastdb',
                      '-in', Path(fasta_path),
                       '-parse_seqids',
                      '-dbtype',dbtype],
                      stdout=PIPE,
                      stderr=PIPE)
        else:
            p = Popen(['makeblastdb',
                      '-in', Path(fasta_path),
                      '-dbtype',dbtype],
                      stdout=PIPE,
                      stderr=PIPE)

        p.wait()
        stdout = str(p.stdout.read())
        stderr = str(p.stderr.read())
        if stderr is not None and stderr != '' and stderr != "b''" :
            logging.error('makeblastdb on {} had the following messages STDERR: {}'.format(fasta_path, stderr))
            return False
        return True




    def run_tblastn(self, query_fasta_path, blast_task, db_path, db_type, min_cov, min_ident, evalue,blast_outfile,logging,num_threads=1,max_target_seqs=100000000,seq_id_file=None):
        if seq_id_file:
            p = Popen(['tblastn',
                       '-query', Path(query_fasta_path),
                       '-seqidlist', '{}'.format(seq_id_file),
                       '-num_threads','{}'.format(num_threads),
                       '-db', '{}'.format(db_path),
                       '-evalue', '{}'.format(evalue),
                       '-out', blast_outfile,
                       '-max_target_seqs','{}'.format(max_target_seqs),
                       '-outfmt', '6 {}'.format(' '.join(BLAST_TABLE_COLS))],
                      stdout=PIPE,
                      stderr=PIPE)
        else:
            p = Popen(['tblastn',
                       '-query', Path(query_fasta_path),
                       '-num_threads','{}'.format(num_threads),
                       '-db', '{}'.format(Path(db_path)),
                       '-evalue', '{}'.format(evalue),
                       '-out', blast_outfile,
                       '-max_target_seqs','{}'.format(max_target_seqs),
                       '-outfmt', '6 {}'.format(' '.join(BLAST_TABLE_COLS))],
                      stdout=PIPE,
                      stderr=PIPE)
        p.wait()
        stdout = str(p.stdout.read())
        stderr = str(p.stderr.read())

        if stdout is not None and stdout != '':
            logging.debug('blastn on db {} and query {} STDOUT: {}'.format(query_fasta_path, db_path, stdout))

        if stderr is None or str(stderr) != '' or str(stderr) != "b''" :
            if os.path.exists(blast_outfile):
                return blast_outfile

        if stderr is not None and (str(stderr) != '' and str(stderr) != "b''" ) :
            logging.debug('blastn on db {} and query {} STDERR: {}'.format(query_fasta_path, db_path, stderr))
            if os.path.exists(blast_outfile):
                return blast_outfile
            else:
                ex_msg = 'tblastn on db {} and query {} did not produce expected output file at {}'.format(
                    query_fasta_path,
                    db_path,
                    blast_outfile)
                logging.error(ex_msg)
                raise Exception(ex_msg)

    def run_blast(self, query_fasta_path, blast_task, db_path, db_type, min_cov, min_ident, evalue,blast_outfile,logging,num_threads=1,word_size=11,max_target_seqs=100000000,seq_id_file=None):

        if seq_id_file :
            p = Popen(['blastn',
                       '-task', blast_task,
                       '-query', Path(query_fasta_path),
                       '-db', '{}'.format(Path(db_path)),
                       '-seqidlist', '{}'.format(seq_id_file),
                       '-num_threads','{}'.format(num_threads),
                       '-evalue', '{}'.format(evalue),
                       '-dust', 'yes',
                       '-perc_identity', '{}'.format(min_ident),
                       '-max_target_seqs','{}'.format(max_target_seqs),
                       '-out', Path(blast_outfile),
                       '-outfmt', '6 {}'.format(' '.join(BLAST_TABLE_COLS))],
                      stdout=PIPE,
                      stderr=PIPE)
        else:
            p = Popen(['blastn',
                       '-task', blast_task,
                       '-query', '{}'.format(Path(query_fasta_path)),
                       '-db', '{}'.format(Path(db_path)),
                       '-num_threads','{}'.format(num_threads),
                       '-evalue', '{}'.format(evalue),
                       '-dust', 'yes',
                       '-perc_identity', '{}'.format(min_ident),
                       '-max_target_seqs','{}'.format(max_target_seqs),
                       '-out', '{}'.format(Path(blast_outfile)),
                       '-outfmt', '6 {}'.format(' '.join(BLAST_TABLE_COLS))],
                      stdout=PIPE,
                      stderr=PIPE)
        p.wait()
        stdout = str(p.stdout.read())
        stderr = str(p.stderr.read())


        if stdout is not None and stdout != '' and str(stdout) != "b''":
            logging.info('blastn on db {} and query {} STDOUT: {}'.format(query_fasta_path, db_path, stdout))

        if stderr is None and stderr != '' and str(stderr) != "b''" :
            if os.path.exists(blast_outfile):
                return blast_outfile

        if stderr is not None and stderr != '' and stderr != "b''":
            logging.error('blastn on db {} and query {} STDERR: {}'.format(query_fasta_path, db_path, stderr))



class BlastReader:
    df = None


    def __init__(self, blast_outfile,logging):
        """Read BLASTN output file into a pandas DataFrame
        Sort the DataFrame by BLAST bitscore.
        If there are no BLASTN results, then no results can be returned.
        Args:
            blast_outfile (str): `blastn` output file path
        Raises:
            EmptyDataError: No data could be parsed from the `blastn` output file
        """
        self.blast_outfile = blast_outfile
        try:
            #self.df = pd.read_table(self.blast_outfile, header=None)
            if not os.path.isfile(blast_outfile):
                logging.warning('No BLASTN results to parse from file %s', blast_outfile)
                self.df = pd.DataFrame()
                return

            if os.path.getsize(blast_outfile) == 0:
                logging.warning('No BLASTN results to parse from file %s', blast_outfile)
                self.df = pd.DataFrame()
                return

            self.df = pd.read_csv(self.blast_outfile,sep="\t",header=None)

            self.df.columns = BLAST_TABLE_COLS

            logger.debug(self.df.head())
            self.is_missing = False


        except EmptyDataError as exc:
            logging.warning('No BLASTN results to parse from file %s', blast_outfile)
            self.is_missing = True
            self.df = pd.DataFrame(index=['A'], columns='A')


    def df_dict(self):
        if not self.is_missing:
            return self.df.to_dict()




#!/usr/bin/env python3
from mob_suite.version import __version__
import os, pycurl, tarfile, zipfile, gzip, multiprocessing, sys
import argparse
import hashlib
import json
import functools
from mob_suite.blast import BlastRunner
from mob_suite.wrappers import mash
import shutil
import datetime
import time #waiting for other processes
from mob_suite.utils import init_console_logger
from mob_suite.constants import  default_database_dir
from ete3 import NCBITaxa
from pathlib import Path

logger = init_console_logger(3)
config_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'config.json')

with open(config_path, 'r') as configfile:
    config = json.load(configfile)

def arguments():


    parser = argparse.ArgumentParser(
        description="MOB-INIT: initialize databases version: {}".format(
            __version__))
    parser.add_argument('-d', '--database_directory',
                        default=default_database_dir,
                        help='Directory to download databases to. Defaults to {}'.format(
                            default_database_dir))

    parser.add_argument('-v', '--verbose',
                        default=0,
                        action='count',
                        help='Set the verbosity level. Can by used multiple times')

    parser.add_argument('-V', '--version', action='version', version="%(prog)s " + __version__ + "")
    args, unknown_args = parser.parse_known_args()
    return args


def check_hash_or_size(filepath, hashsums):
    """
    Calculate the SHA256 of the given file and compare it to the known good
    value. Returns True if the hashes match, and False if the file is missing or
    the hashes do not match.
    Otherwise uses minimum file size of >= 215 MB
    FigShare mirror returns non hash identical files for some reason

    :param filepath: Path to the file
    :param hashsums:  Expected SHA256 hashes (given different database versions)
    :return: True if the hashes match, False otherwise
    """

    try:
        with open(filepath, 'rb') as f:
            sha = hashlib.sha256()
            contents = f.read()
            sha.update(contents)

    except FileNotFoundError:
        return False

    h_obs = sha.hexdigest()
    logger.info("Download data.zip sha256 checksum is {}".format(h_obs))
    logger.info("Download data.zip size in bytes is {}".format(os.path.getsize(filepath)))
    hash_check_bool = any([refhash == h_obs for refhash in hashsums])
    size_check_bool = (os.path.getsize(filepath) >= 225377917)
    return any([hash_check_bool, size_check_bool])


def download_to_file(url, file):
    """
    Downloads a file from a URL using curl.

    :param url:  Source URL from which the file is downloaded
    :param file: The destination file
    :return: None
    """

    with open(file, 'wb') as f:
        c = pycurl.Curl()
        c.setopt(c.URL, url)
        # Follow redirect.
        c.setopt(c.FOLLOWLOCATION, True)
        c.setopt(c.WRITEDATA, f)
        c.setopt(c.NOPROGRESS, False)
        c.perform()
        c.close()


def extract(fname, outdir):
    """
    Decompress a zip or gzip archive. Decompression method is selected based
    on file extension. Following extraction, the original archive is deleted.

    :param fname:  Path to the archive to be extracted
    :param outdir: Directory into which the results are placed
    :return: None
    """

    logger.info(f'Decompressing {fname}')
    shutil.unpack_archive(fname, outdir)
    dir_name = os.path.join(outdir,os.path.basename(fname))
    for ext in ['.tar.gz','.zip','.gz']:
        dir_name = dir_name.replace(ext,'')

    src_files = os.listdir(dir_name)
    for file_name in src_files:
        full_file_name = os.path.join(dir_name, file_name)
        if os.path.isfile(full_file_name):
            shutil.copyfile(full_file_name, os.path.join(outdir,file_name))
    shutil.rmtree(dir_name)
    os.remove(fname)

def main():
    
    args = arguments()
    

    database_directory = os.path.abspath(args.database_directory)


    if os.path.exists(database_directory) == False:
        os.makedirs(database_directory)
        logger.info("Database directory folder created at {}".format(database_directory))
    else:
        logger.info("Database directory folder already exists at {}".format(database_directory))

    
    # Helper function to simplify adding database_directory to everything
    prepend_db_dir = functools.partial(os.path.join, database_directory)

    lockfilepath=os.path.join(database_directory,".lock")
    status_file = prepend_db_dir('status.txt')

    if os.path.exists(lockfilepath) == False:
        try:
            open(file=lockfilepath, mode="w").close()
            logger.info("Placed lock file at {}".format(lockfilepath))
        except Exception as e:
            logger.error("Failed to place a lock file at {}. Database diretory can not be accessed. Wrong path?".format(lockfilepath))
            logger.error("{}".format(e))
            pass
    else:
        while os.path.exists(lockfilepath):
            elapsed_time = time.time() - os.path.getmtime(lockfilepath)
            logger.info("Lock file found at {}. Waiting for other processes to finish database init ...".format(lockfilepath))
            logger.info("Elapsed time {} min. Will continue processing after 16 min mark.".format(int(elapsed_time/60)))
            if elapsed_time >= 1000:
                logger.info("Elapsed time {} min. Assuming previous process completed all init steps. Continue ...".format(int(elapsed_time/60)))
                try: #if previous process failed, no processes are running and > 16 min passed since the lock was created
                    os.remove(lockfilepath)
                except: #continue if file was removed by other process
                    pass
                break
            time.sleep(60) #recheck every 1 min if lock file was removed
        logger.info("Lock file no longer exists. Assuming init process completed successfully")
        return 0



    logger.info('Initializing databases...this will take some time')
    # Find available threads and use the maximum number available for mash sketch but cap it at 32
    try:
        num_threads = len(os.sched_getaffinity(0))
    except AttributeError:
        num_threads = multiprocessing.cpu_count()

    if num_threads > 32:
        num_threads = 32
    if num_threads < 1:
        num_threads = 1

    if not os.path.exists(database_directory):
        os.makedirs(database_directory)

    zip_file = prepend_db_dir('data.tar.gz')
    plasmid_database_fasta_file = prepend_db_dir('ncbi_plasmid_full_seqs.fas')
    repetitive_fasta_file = prepend_db_dir('repetitive.dna.fas')
    mash_db_file =  prepend_db_dir('ncbi_plasmid_full_seqs.fas.msh')

    logger.info('Downloading databases...this will take some time')

    for db_mirror in config['db_mirrors']:
        try:
            logger.info('Trying mirror {}'.format(db_mirror))
            download_to_file(db_mirror, zip_file)
            break
        except Exception as e:
            logger.error("Download failed with error {}. Removing lock file".format(str(e)))
            os.remove(lockfilepath)
            sys.exit(-1)


    logger.info("Downloading databases successful, now building databases at {}".format(database_directory))
    extract(zip_file, database_directory)

    files = [prepend_db_dir(f)
             for f in os.listdir(database_directory)
             if f.endswith('.gz')]

    for file in files:

        extract(file, database_directory)

    #Initialize blast and mash databases
    try:
        logger.info('Building repetitive mask database')
        blast_runner = BlastRunner(repetitive_fasta_file, database_directory)
        blast_runner.makeblastdb(repetitive_fasta_file, 'nucl',logger)

        logger.info('Building complete plasmid database')
        blast_runner = BlastRunner(plasmid_database_fasta_file, database_directory)
        blast_runner.makeblastdb(plasmid_database_fasta_file, 'nucl',logger,True)

        logger.info('Sketching complete plasmid database')
        mObj = mash()
        mObj.mashsketch(plasmid_database_fasta_file,
                        mash_db_file,
                        num_threads=num_threads)
    except Exception as e:
        logger.error('Downloading databases failed, please check your internet connection and retry')
        logger.error("Process failed with error {}. Removing lock file".format(e))
        os.remove(lockfilepath)
        sys.exit(-1)

    try:
        logger.info("Init ete3 library ...")
        ete3taxadbpath = os.path.abspath(os.path.join(database_directory,"taxa.sqlite"))
        NCBITaxa(dbfile=ete3taxadbpath) #the creatuib if NCBITaxa class triggers update_taxonomy_database()
    except Exception as e:
        logger.error("Init of ete3 library failed with error {}. Removing lock file".format(e))
        os.remove(lockfilepath)
        sys.exit(-1)

    try:
        os.remove(os.path.join(os.getcwd(), "taxdump.tar.gz"))
        logger.info("Removed residual taxdump.tar.gz as ete3 is not doing proper cleaning job.")
    except:
        pass

    with open(status_file, 'w') as f:
        download_date = datetime.datetime.today().strftime('%Y-%m-%d')
        f.write("Download date: {}. Removing lock file.".format(download_date))
        try:
            os.remove(lockfilepath)
        except:
            logger.warning("Lock file is already removed by some other process.")
            pass


    logger.info("MOB init completed successfully")
    return 0

# call main function
if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import os
from distutils.core import setup
from setuptools import find_packages
from mob_suite.version import __version__
author = 'James Robertson, Kyrylo Bessonov'

classifiers = """
Development Status :: 3 - Alpha
Environment :: Console
License :: OSI Approved :: Apache Software License
Intended Audience :: Science/Research
Topic :: Scientific/Engineering
Topic :: Scientific/Engineering :: Bio-Informatics
Programming Language :: Python
Programming Language :: Python :: 3.7
Programming Language :: Python :: 3.8
Programming Language :: Python :: Implementation :: CPython
Operating System :: POSIX :: Linux
""".strip().split('\n')


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


exec(open('mob_suite/version.py').read())

setup(
    name='mob_suite',
    include_package_data=True,
    version=__version__,
    python_requires='>=3.7.0,<4',
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    packages=find_packages(exclude=['tests', 'databases']),
    url='https://github.com/phac-nml/mob-suite',
    license='GPLv3',
    author='James Robertson',
    author_email='james.robertson@canada.ca',
    description=(
        'MOB-suite is a set of tools for finding, typing and reconstruction of plasmids from draft and complete genome assemblies.'),
    keywords='Plasmids finding typing reconstruction',
    classifiers=classifiers,
    package_dir={'mob_suite': 'mob_suite'},
    package_data={'mob_suite': ['config.json']},

    install_requires=[
        'numpy>=1.11.1,<1.23.5',
        'tables>=3.3.0,<4',
        'pandas>=0.22.0,<=1.0.5',
        'biopython>=1.8,<2',
        'pycurl>=7.43.0,<8',
        'scipy>=1.1.0,<2',
        'ete3>=3.1.3,<4',
        'six>=1.10,<2',
    ],

    entry_points={
        'console_scripts': [
            'mob_init=mob_suite.mob_init:main',
            'mob_recon=mob_suite.mob_recon:main',
            'mob_cluster=mob_suite.mob_cluster:main',
            'mob_typer=mob_suite.mob_typer:main',
        ],
    },
)
